# Loonies 8192 r50 for DOS and SDL platforms
# Copyright (C) 2018-2020 Thomas Perl <m@thp.io>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import yaml
import os
from collections import defaultdict

LSTRINGS_YAML = os.path.normpath(os.path.join(os.path.dirname(__file__), 'input', 'lstrings.yaml'))
LSTRINGS_DE_YAML = os.path.normpath(os.path.join(os.path.dirname(__file__), 'input', 'lstrings.de.yaml'))
LSTRINGS_H = os.path.normpath(os.path.join(os.path.dirname(__file__), '..', '..', 'src', 'gen', 'lstrings.h'))
LSTRINGS_CPP = os.path.normpath(os.path.join(os.path.dirname(__file__), '..', '..', 'src', 'gen', 'lstrings.cpp'))

init_commands = defaultdict(list)
databuffer = b''
datakey = b'\xca\xfe\xf0\x0d\xfa\xce'

def dumpbuffer(buf, name, fp):
    print(f'unsigned char {name}[{len(buf)}] =', file=fp)
    print('{', file=fp)
    for idx, ch in enumerate(buf):
        print(f'0x{ch:02x}', end=', ', file=fp)
        if idx % 16 == 15:
            print(file=fp)
    print('};', file=fp)


def offset_of_string(stringdata):
    global databuffer

    stringdata = stringdata.encode('ascii') + b'\0'

    try:
        return databuffer.index(stringdata)
    except ValueError:
        result = len(databuffer)
        databuffer += stringdata
        return result


def process(yaml_file, language_index):
    global init_commands

    for k, v in yaml.load(open(yaml_file).read()).items():
        if isinstance(v, str):
            if language_index == 0:
                print(f'extern const char *{k};', file=hfp)
                print(f'const char *{k};', file=cfp)
            init_commands[language_index].append(f'{k} = (const char *)buf+{offset_of_string(v)};')
        elif isinstance(v, list):
            if language_index == 0:
                print(f'extern const char *{k}[];', file=hfp)
                print(f'const char *{k}[{len(v)}];', file=cfp)
            for idx, item in enumerate(v):
                if item is None:
                    init_commands[language_index].append(f'{k}[{idx}] = 0;')
                else:
                    init_commands[language_index].append(f'{k}[{idx}] = (const char *)buf+{offset_of_string(item)};')
        else:
            raise ValueError(v)


print(f'Writing to {LSTRINGS_H} and {LSTRINGS_CPP}')
with open(LSTRINGS_H, 'w', encoding='utf-8') as hfp, open(LSTRINGS_CPP, 'w', encoding='utf-8') as cfp:
    print('void lstrings_init();', file=hfp)
    print('void lstrings_choose(int language);', file=hfp)

    process(LSTRINGS_YAML, 0)
    process(LSTRINGS_DE_YAML, 1)

    databuffer = bytes(ch ^ datakey[i%len(datakey)] for i, ch in enumerate(databuffer))
    dumpbuffer(databuffer, 'buf', cfp)
    dumpbuffer(datakey, 'key', cfp)
    print('void lstrings_init() {', file=cfp)
    print('    static int inited=0; if (inited) { return; } inited = 1;', file=cfp)
    print('    int i;', file=cfp)
    print(f'    for (i=0; i<{len(databuffer)}; ++i)', file=cfp)
    print(f'        buf[i] ^= key[i%{len(datakey)}];', file=cfp)
    print('}', file=cfp)
    print('void lstrings_choose(int language) {', file=cfp)
    for k, v in init_commands.items():
        print('    if (language == %s) {' % (k,), file=cfp)
        for command in v:
            print(f'        {command}', file=cfp)
        print('    }', file=cfp)
    print('}', file=cfp)
