# Loonies 8192 r50 for DOS and SDL platforms
# Copyright (C) 2018-2020 Thomas Perl <m@thp.io>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.


# Example usage: python3 layout.py --load ../../src/game_layout_dos.h
#
# Interactive layout editor + subdivision into non-overlapping rects
# so that the renderer can fill the screen without overdraw. Can load
# header files it generates for editing again (only base definition).
#
# The "align_to" variable can be used for pixel snapping, so that the
# DOS VGA rendering routines can reliably work on 4-pixel boundaries
# (theoretically only needed for the X axis, but enforced for Y too).

import pygame
from pygame.locals import *

import re
import argparse
import unittest

debug = False
align_to = 4

KIND_DARK, KIND_LIGHT, KIND_PLAYFIELD = range(3)

class Rect():
    def __init__(self, x, y, w, h, kind, desc=''):
        self.x = x
        self.y = y
        self.w = w
        self.h = h
        self.kind = kind
        self.desc = desc
        self.path = desc
        self.do_not_align_to_grid = False

    def __repr__(self):
        return f'<{self.__class__.__name__}(x={self.x}, y={self.y}, w={self.w}, h={self.h}, kind={self.kind}, desc={self.desc!r}>'

    def __eq__(self, other):
        return type(self) == type(other) and self.__dict__ == other.__dict__

    def info(self):
        return (int(self.x), int(self.y), int(self.w), int(self.h), self.kind, self.desc)

    def set_W(self, value):
        self.x = self.w + self.x - value
        self.w = value

    def get_W(self):
        return self.w

    W = property(fget=get_W, fset=set_W)

    def set_H(self, value):
        self.y = self.h + self.y - value
        self.h = value

    def get_H(self):
        return self.h

    H = property(fget=get_H, fset=set_H)

    def contains(self, x, y):
        return x >= self.x and x < self.x + self.w and y >= self.y and y < self.y + self.h

    def draw(self, recurse=True):
        #pygame.draw.rect(out, (90, 90, 90), (self.x, self.y, self.w, self.h), 1)
        pygame.draw.line(out, (90, 90, 90), (self.x,          self.y),          (self.x + self.w, self.y))
        pygame.draw.line(out, (90, 90, 90), (self.x,          self.y + self.h), (self.x + self.w, self.y + self.h))
        pygame.draw.line(out, (90, 90, 90), (self.x,          self.y),          (self.x,          self.y + self.h))
        pygame.draw.line(out, (90, 90, 90), (self.x + self.w, self.y),          (self.x + self.w, self.y + self.h))
        if recurse:
            self.topleft().draw(False)
            self.bottomright().draw(False)

    def text(self, msg):
        surf = fnt.render(msg, True, (200, 200, 200))
        w, h = surf.get_size()
        out.blit(surf, (self.x + (self.w - w) / 2, self.y + (self.h - h) / 2))

    def topleft(self):
        return type(self)(self.x-3, self.y-3, 7, 7, self.kind)

    def bottomright(self):
        return type(self)(self.x+self.w-3, self.y+self.h-3, 7, 7, self.kind)

    def with_metrics(self, txt):
        return f'{txt} ({self.w:.0f}x{self.h:.0f} @ {self.x:.0f},{self.y:.0f})'

    def fixup(self):
        if self.w < 0:
            self.x += self.w
            self.w *= -1
        if self.h < 0:
            self.y += self.h
            self.h *= -1

        if self.do_not_align_to_grid:
            return

        for axis in 'xywh':
            diff = getattr(self, axis) % align_to
            if diff != 0:
                if diff <= (align_to / 2):
                    setattr(self, axis, getattr(self, axis) - diff)
                else:
                    diff = align_to - diff
                    setattr(self, axis, getattr(self, axis) + diff)


def load_layout(s):
    match = re.findall(r'struct GameLayout {(?:.*)} game_layout = {(.*?)};', s, re.MULTILINE | re.DOTALL)[0]
    for line in match.splitlines():
        line = line.strip()
        if line:
            x, y, w, h, desc, kind = re.match(r'{ (\d+), (\d+), (\d+), (\d+), (\w+), (\d+) },', line).groups()
            x, y, w, h, kind = int(x), int(y), int(w), int(h), int(kind)
            desc = desc.lower()[len('game_layout_'):]
            rect = Rect(x, y, w, h, kind, desc)
            if rect.x % align_to != 0 or rect.y % align_to != 0 or rect.w % align_to != 0 or rect.h % align_to != 0:
                rect.do_not_align_to_grid = True
            yield rect


class TestLoadLayout(unittest.TestCase):
    def test_load_layout(self):
        r = list(load_layout("""
        struct GameLayout {
            LayoutRect bg;
            LayoutRect bgfix;
            LayoutRect title_bar;
            LayoutRect play_field;
            LayoutRect score_1;
            LayoutRect score_2;
            LayoutRect score_3;
            LayoutRect block;
        } game_layout = {
            { 0, 0, 400, 240, GAME_LAYOUT_BG, 1 },
            { 292, 84, 100, 148, GAME_LAYOUT_BGFIX, 1 },
            { 272, 8, 120, 52, GAME_LAYOUT_TITLE_BAR, 0 },
            { 8, 8, 256, 224, GAME_LAYOUT_PLAY_FIELD, 2 },
            { 292, 84, 100, 44, GAME_LAYOUT_SCORE_1, 0 },
            { 292, 136, 100, 44, GAME_LAYOUT_SCORE_2, 0 },
            { 292, 188, 100, 44, GAME_LAYOUT_SCORE_3, 0 },
            { 1000, 1000, 16, 16, GAME_LAYOUT_BLOCK, 2 },
        };
        """))
        self.assertEqual(r, [
            Rect(0, 0, 400, 240, 1, 'bg'),
            Rect(292, 84, 100, 148, 1, 'bgfix'),
            Rect(272, 8, 120, 52, 0, 'title_bar'),
            Rect(8, 8, 256, 224, 2, 'play_field'),
            Rect(292, 84, 100, 44, 0, 'score_1'),
            Rect(292, 136, 100, 44, 0, 'score_2'),
            Rect(292, 188, 100, 44, 0, 'score_3'),
            Rect(1000, 1000, 16, 16, 2, 'block'),
        ])


def put_above(above, below):
    a = Rect(below.x, below.y, below.w, (above.y - below.y), below.kind, below.desc)
    a.path = below.path + '/a'
    if a.h > below.h:
        a.h = below.h
    b = Rect(below.x, above.y, (above.x - below.x), above.h, below.kind, below.desc)
    b.path = below.path + '/b'
    if b.y < below.y:
        diff = (below.y - b.y)
        b.y += diff
        b.h -= diff
    elif b.y + b.h > below.y + below.h:
        diff = (b.y + b.h) - (below.y + below.h)
        b.h -= diff
    if b.w > below.w:
        b.w = below.w
    if b.h > below.h:
        b.h = below.h
    c = Rect(above.x, above.y, above.w, above.h, above.kind, above.desc)
    c.path = above.path + '/c'
    d = Rect(above.x + above.w, above.y, (below.x + below.w) - (above.x + above.w), above.h, below.kind, below.desc)
    d.path = below.path + '/d'
    if d.y < below.y:
        diff = (below.y - d.y)
        d.y += diff
        d.h -= diff
    elif d.y + d.h > below.y + below.h:
        diff = (d.y + d.h) - (below.y + below.h)
        d.h -= diff
    if d.x < below.x:
        d.x = below.x
        d.w = below.w
    if d.h > below.h:
        d.h = below.h
    e = Rect(below.x, above.y + above.h, below.w, (below.y + below.h) - (above.y + above.h), below.kind, below.desc)
    e.path = below.path + '/e'
    if e.y < below.y:
        diff = (below.y - e.y)
        e.y += diff
        e.h -= diff
    for n in 'abcde':
        r = locals()[n]
        if r.w > 0 and r.h > 0:
            yield r


def merge_rects(rects):
    merged = True
    while merged:
        merged = False
        for i, a in enumerate(rects):
            for j, b in enumerate(rects[i+1:]):
                if a.kind == b.kind and a.x == b.x and a.w == b.w and a.y + a.h == b.y:
                    a.h = a.h + b.h
                    b.w = 0
                    merged = True
                    break
            if merged:
                break

    last = len(rects)-1
    for j in range(len(rects)):
        j = last-j
        for i in range(j):
            if rects[i].info() == rects[j].info():
                rects.pop(j)
                break

    return [r for r in rects if r.w > 0]


def subdivide(rects):
    todo = list(rects)
    temp = []

    while todo:
        above = todo.pop(0)
        new_temp = []

        split = False
        for below in temp:
            new_temp.extend(put_above(above, below))
            split = True

        if not split:
            new_temp.append(above)

        temp = new_temp

    return merge_rects(temp)


def save(filename, rects):
    print(f'Saving {rects} to {filename}')
    bg_rect = next(rect for rect in rects if rect.desc == 'bg')
    with open(filename, 'w') as fp:
        print('/* Exported/generated from layout.py */', file=fp)
        print(f'#define PORT_WIDTH {bg_rect.w}', file=fp)
        print(f'#define PORT_HEIGHT {bg_rect.h}', file=fp)
        print("""
struct LayoutRect {
    unsigned short x, y, w, h;
    unsigned short name;
    unsigned short kind;
};
        """.strip(), file=fp)
        print('enum GameLayoutRectName {', file=fp)
        for rect in rects:
            print(f'    GAME_LAYOUT_{rect.desc.upper()},', file=fp)
        print('};', file=fp)
        print('struct GameLayout {', file=fp)
        for rect in rects:
            print(f'    LayoutRect {rect.desc};', file=fp)
        print('} game_layout = {', file=fp)
        for rect in rects:
            print('    { %d, %d, %d, %d, GAME_LAYOUT_%s, %d },' % (rect.x, rect.y, rect.w, rect.h, rect.desc.upper(), rect.kind), file=fp)
        print('};', file=fp)
        print('struct LayoutRect subdivided[] = {', file=fp)
        for rect in sorted(subdivide(rects), key=lambda rect: (rect.y, rect.x, rect.w, rect.h, rect.kind, rect.desc)):
            print('    { %d, %d, %d, %d, GAME_LAYOUT_%s, %d },' % (rect.x, rect.y, rect.w, rect.h, rect.desc.upper(), rect.kind), file=fp)
        print('};', file=fp)


if __name__ == '__main__':
    pygame.font.init()

    parser = argparse.ArgumentParser(description='Rectangle layouting')
    parser.add_argument('--load', metavar='FILENAME', type=str, default=None)
    args = parser.parse_args()
    print(args)

    fnt = pygame.font.Font(pygame.font.get_default_font(), 10)

    rects = [
        Rect(0, 0, 400, 240, KIND_LIGHT, 'bg'),
        Rect(168, 36, 144, 140, KIND_LIGHT, 'bgfix'),
        Rect(0, 0, 320, 30, KIND_DARK, 'title_bar'),
        Rect(4, 36, 16*10, 14*10, KIND_PLAYFIELD, 'play_field'),
        Rect(170, 38, 145, 14, KIND_DARK, 'score_1'),
        Rect(170, 55, 145, 14, KIND_DARK, 'score_2'),
        Rect(170, 164, 145, 14, KIND_DARK, 'score_3'),
        Rect(1000, 1000, 16, 16, KIND_PLAYFIELD, 'block'),
    ]
    rects[-1].do_not_align_to_grid = True

    if args.load:
        rects = list(load_layout(open(args.load).read()))

    g_width = 320
    g_height = 240
    for rect in rects:
        if rect.desc == 'bg':
            g_width = rect.w
            g_height = rect.h
            break

    out = pygame.Surface((g_width, g_height), 0, 32)
    offset_x, offset_y = 20, 20
    scale = 2

    screen = pygame.display.set_mode((2*offset_x+g_width*2, 2*offset_y+g_height*scale))

    pygame.display.set_caption(f'{g_width}x{g_height} ({args.load or "<default>"}) a=add, s=save, p=subdiv, d=dump, q=quit')

    subdivided = []
    grabbed = (None, None, None)
    quit = False
    while not quit:
        out.fill((30, 30, 30))

        for x in range(0, g_width, align_to):
            pygame.draw.line(out, (40, 40, 40), (x, 0), (x, g_height))

        for y in range(0, g_height, align_to):
            pygame.draw.line(out, (40, 40, 40), (0, y), (g_width, y))

        if grabbed == (None, None, None):
            for r in rects:
                r.fixup()

        if grabbed != (None, None, None) or not subdivided:
            for idx, rect in enumerate(rects):
                rect.text(rect.with_metrics(rect.desc))
                rect.draw()

        if grabbed == (None, None, None):
            for part in subdivided:
                desc = part.desc if debug else (part.desc[:part.desc.index('/')] if '/' in part.desc else part.desc)
                part.text(part.with_metrics(desc))
                part.draw()

        screen.blit(pygame.transform.scale(out, (g_width*scale, g_height*scale)), (offset_x, offset_y))
        pygame.display.flip()

        while True:
            e = pygame.event.poll()
            if e.type == NOEVENT:
                break
            elif e.type == QUIT:
                quit = True
                break
            elif e.type == KEYDOWN:
                if e.key == K_d:
                    print(rects)
                elif e.key == K_q:
                    quit = True
                    break
                elif e.key == K_a:
                    rects.append(Rect(10, 10, 100, 100, len(rects)))
                elif e.key == K_s:
                    save(args.load or 'game_layout.h', rects)
                elif e.key == K_p:
                    subdivided = subdivide(rects)
            elif e.type == MOUSEBUTTONDOWN:
                # first check for border drag
                pos = ((e.pos[0] - offset_x) / scale, (e.pos[1] - offset_y) / scale)
                for rect in rects:
                    if rect.bottomright().contains(*pos):
                        grabbed = (rect, 'wh', +1)
                    elif rect.topleft().contains(*pos):
                        grabbed = (rect, 'WH', -1)
                # if nothing found, check for center grab
                if grabbed == (None, None, None):
                    for rect in rects:
                        if rect.contains(*pos):
                            grabbed = (rect, 'xy', +1)
            elif e.type == MOUSEBUTTONUP:
                grabbed = (None, None, None)
                subdivided = []
            elif e.type == MOUSEMOTION:
                if grabbed != (None, None, None):
                    rect, (xname, yname), multiplier = grabbed
                    setattr(rect, xname, getattr(rect, xname) + e.rel[0] * multiplier / scale)
                    setattr(rect, yname, getattr(rect, yname) + e.rel[1] * multiplier / scale)

