/*
    Loonies 8192 r50 for DOS and SDL platforms
    Copyright (C) 2018-2020 Thomas Perl <m@thp.io>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
*/


/**
 * LuminDOS
 * 2018-06-18/19 Thomas Perl <m@thp.io>
 * Based on the Python implementation 2018-06-10
 **/

#include "loonport.h"

// Assuming a double-buffered screen, after the first two screen pages we are
// "offscreen" and can put data there (blit buffer)
#define OFFSCREEN_Y (PORT_HEIGHT*2)

// Audio related data + MusicPlayer implementation
#include "data/midifreq.h"
#include "audio.h"

// Background music data
#include "gen/music.h"

// Time step
#include "timestep.h"

// Game logic
#include "gamelogic.h"

// Music data:
//  - Raw OPL-2 command stream [_out]
//  - Raw MIDI command stream [_dim]
//  - CD Audio Track number (e.g. 2 or 3)
//  - MIDI channel where the melody plays (e.g. 1 or 0) -- used for PC speaker
MusicDef MUSICDEF[2] = {
    {{{samoff_out, samoff_out_len}, {samoff_dim, samoff_dim_len}}, 2, 1},
    {{{clickedy_out, clickedy_out_len}, {clickedy_dim, clickedy_dim_len}}, 3, 0},
};

// Color schemes, block (tile) data and digit data
#include "gen/schemes.h"

// Graphics artwork (bomb animation)
#include "gen/graphics.h"

// Raw font data (all text except score digits)
#include "data/font.h"

// Localization strings
#include "gen/lstrings.h"
#include "gen/lstrings.cpp"

#if defined(BUILD_FOR_3DS) || defined(BUILD_FOR_PSP)
#define AUTOCONFIGURE_AUDIO_DEVICE
#endif

// Convenience function: Clear the first screen page
void vgaclear() { vgaclearbytes((long)PORT_WIDTH * (long)PORT_HEIGHT); }

// Global palette rotation index (for animation fx)
int palrotate = 0;

// Default configuration
static struct HardwareConfig
hw_cfg = {
    LANGUAGE_ENGLISH,
    AUDIO_DEV_NONE,
    AUDIO_OPL_ADLIB_388,
    AUDIO_MPU_330,
    SOUND_TRACK_DESCENT,
    0,
};

bool is_back(int key) { return (key == KeyCode_q || key == KeyCode_Escape); }
bool is_enter(int key) { return (key == KeyCode_Space || key == KeyCode_Enter); }
bool is_up(int key) { return key == KeyCode_k || key == KeyCode_Up; }
bool is_down(int key) { return key == KeyCode_j || key == KeyCode_Down; }
bool is_left(int key) { return key == KeyCode_h || key == KeyCode_Left; }
bool is_right(int key) { return key == KeyCode_l || key == KeyCode_Right; }

static int
mystrlen(const char *str)
{
    int len = 0;
    while (str && str[len]) {
        ++len;
    }
    return len;
}

// "diamond" is the pattern used to draw the background, there are
// multiple "modes" that correspond to different patterns
static const int MAXDIAMOND = 6;
int diamondmode = 0;

struct SaveGame {
    void set(score_t score);
    bool verify();

    score_t score;
    score_t checksum;
};

void
SaveGame::set(score_t score)
{
    this->score = score;
    this->checksum = score ^ SCORE_XOR_VALUE;
}

bool
SaveGame::verify()
{
    return (checksum == (score ^ SCORE_XOR_VALUE));
}

// Unpacked pixel data for the score digits (0-9)
static unsigned char
digit_pixels[10][10*10];

// Current tile set in use (in-game blocks)
int tileset = 0;

// Unpacked pixel data for the current tile set
static unsigned char
tile_pixels[16*16];

// Unpack a 2-bit-per-pixel encoded bitmap into 1-byte-per-pixel data
static void
unpack_2bit(unsigned char *src, unsigned char *dst, int size)
{
    int max = size*size*2/8;
    for (int i=0; i<max; ++i) {
        unsigned char v = *src++;
        *dst++ = (v >> 6) & 3;
        *dst++ = (v >> 4) & 3;
        *dst++ = (v >> 2) & 3;
        *dst++ = v & 3;
    }
}

// Load all packed score digits (0-9) for usage -- 10x10 pixels
static void
load_digits()
{
    for (int i=0; i<10; ++i) {
        unpack_2bit(packed_digits[i].pixels_2bit, digit_pixels[i], 10);
    }
}

// Load the tile set identified by "ts" into the tile pixels -- 16x16 pixels
static void
load_tiles(int ts)
{
    unpack_2bit(packed_tiles[ts].pixels_2bit, tile_pixels, 16);
}

// Render text onto the screen (using the 8x8 font from data/font.h),
// optionally cutting off text rendering after "n" characters (default: until '\0')
static void
render_font(int x, int y, const char *msg, unsigned char color, int n=-1)
{
    for (int ch=0; msg[ch] && (n==-1 || ch < n); ch++) {
        int c = msg[ch];
        for (int xx=0; xx<8; xx++) {
            if (x+8*ch+xx >= PORT_WIDTH) {
                continue;
            }
            int offset = (c-32)*8+xx;
            if (offset > out_font_dat_len) {
                continue;
            }
            unsigned char col = out_font_dat[offset];
            for (int yy=0; yy<8; yy++) {
                if (y+yy >= PORT_HEIGHT*4/*VGA pages */) {
                    continue;
                }
                if (col & (1 << yy)) {
                    // TODO: Make use of planar mode and avoid plane switching
                    putpixel(x + 8 * ch + xx, y+yy, color);
                }
            }
        }
    }
}

// Helper function for line_split() below -- search for suitable break
int line_split_step(const char **s, int width)
{
    int take = mystrlen(*s);
    if (take > width) {
        const char *end = (*s) + width - 1;
        // Search for a suitable break
        while (end > *s && *end != ' ') {
            --end;
        }
        if (end == *s) {
            take = width;
        } else {
            take = end - *s + 1;
        }
    }
    *s += take;
    return take;
}

// Split a string into lines with a maximum text length, call the callback for
// each line and optionally strip spaces at the beginning of each string
int line_split(const char *input, int textwidth, void (*cb)(const char *start, int len, void *user_data), void *user_data, bool strip_spaces=true)
{
    int lines = 0;
    const char *saveptr = input;

    if (!*saveptr) {
        if (cb) {
            cb(input, 0, user_data);
        }
        return 1;
    }

    while (*saveptr) {
        while (strip_spaces && *saveptr && *saveptr == ' ') {
            ++saveptr;
        }
        if (!*saveptr) {
            break;
        }
        const char *start = saveptr;
        int len = line_split_step(&saveptr, textwidth);
        if (cb) {
            cb(start, len, user_data);
        }
        lines++;
    }
    return lines;
}

struct RenderFontWrapCtx {
    int x;
    int y;
    int color;
};

// Helper function for below
void
render_font_wrap_line(const char *line, int n, void *user_data)
{
    struct RenderFontWrapCtx *ctx = (struct RenderFontWrapCtx *)user_data;

    render_font(ctx->x, ctx->y, line, ctx->color, n);

    ctx->y += 9;
}

// Render text with line wrapping (taking care of line splitting, state mgmt, etc...)
static void
render_font_wrap(int x, int y, const char *msg, unsigned char color, int textwidth)
{
    struct RenderFontWrapCtx ctx;
    ctx.x = x;
    ctx.y = y;
    ctx.color = color;

    line_split(msg, textwidth, render_font_wrap_line, &ctx);
}

// Render font, but clip at X and Y coordinates, used for the scrolling labels in the options menu
static void
render_font_xclipped(int x, int y, const char *msg, unsigned char color, int x0, int x1, int n=-1)
{
    for (int ch=0; msg[ch] && (n==-1 || ch < n); ch++) {
        int c = msg[ch];
        for (int xx=0; xx<8; xx++) {
            int xp = x+8*ch+xx;
            if (xp >= PORT_WIDTH || xp < x0 || xp > x1) {
                continue;
            }
            int offset = (c-32)*8+xx;
            if (offset > out_font_dat_len) {
                continue;
            }
            unsigned char col = out_font_dat[offset];
            for (int yy=0; yy<8; yy++) {
                if (y+yy >= PORT_HEIGHT*4/*VGA pages */) {
                    continue;
                }
                if (col & (1 << yy)) {
                    // TODO: Make use of planar mode and avoid plane switching
                    putpixel(x + 8 * ch + xx, y+yy, color);
                }
            }
        }
    }
}

// Render font and scale with a integer scaling factor "N" (used for screen titles)
void
render_font_x(int x, int y, const char *msg, char color, int N)
{
    for (int ch=0; msg[ch]; ch++) {
        int c = msg[ch];
        for (int xx=0; xx<8; xx++) {
            int offset = (c-32)*8+xx;
            if (offset > out_font_dat_len) {
                continue;
            }
            unsigned char col = out_font_dat[offset];
            for (int yy=0; yy<8; yy++) {
                if (col & (1 << yy)) {
                    for (int xxx=0; xxx<N; xxx++) {
                        for (int yyy=0; yyy<N; yyy++) {
                            // TODO: Make use of planar mode and avoid plane switching
                            putpixel(x + (8 * ch + xx)*N + xxx, y+yy*N+yyy, color);
                        }
                    }
                }
            }
        }
    }
}

// Prepare numbers for fast VGA blitting by rendering them to the designated offscreen area
static void
prepare_number()
{
    unsigned char mapping[4];
    mapping[0] = 0;

    for (int color=1; color<=2; ++color) {
        mapping[1] = color;
        mapping[2] = color + 2;
        mapping[3] = color + 4;
        for (int digit=0; digit<=9; ++digit) {
            for (int yy=0; yy<10; ++yy) {
                for (int xx=0; xx<10; ++xx) {
                    unsigned int value = digit_pixels[digit][yy*10+xx];
                    if (value) {
                        putpixel(6*game_layout.block.w + xx + 12 * digit, OFFSCREEN_Y - 20 + 10*(color-1) + yy, mapping[value]);
                    }
                }
            }
        }
    }
}

// Slow per-pixel blitting function that treats color 0 as transparent
void
transparentblit(int sx, int sy, int dx, int dy, int w, int h)
{
    for (int yy=0; yy<h; ++yy) {
        for (int xx=0; xx<w; ++xx) {
            int ch = getpixel(sx+xx, sy+yy);
            if (ch != 0) {
                putpixel(dx+xx, dy+yy, ch);
            }
        }
    }
}

// Render a score number by blitting the 10x10 font digits
static void
render_number(int x, int y, score_t number, unsigned char color)
{
    do {
        int digit = number % 10;
#if defined(__BORLANDC__)
        // Use fast VGA blitting for DOS (no transparency, though..)
        vgablit(6*game_layout.block.w+12*digit, OFFSCREEN_Y - 20 + 10*(color-1), x, y, 10, 10, 0x00);
#else
        // Everywhere else use the slower transparent blitting
        transparentblit(6*game_layout.block.w+12*digit, OFFSCREEN_Y - 20 + 10*(color-1), x, y, 10, 10);
#endif
        number /= 10;
        x -= 8;
    } while (number > 0);
}

static void busy_wait_ms(time_ms_t duration)
{
    vgawait();
    time_ms_t then = time_now_ms() + duration;
    while (time_now_ms() < then);
}

// Current RGB color palette for background color rotation (64 entries)
static unsigned char rgb[64][3];

// update the background color palette applying an offset (color cycling)
void vgapalette(int offset)
{
    for (int i=0; i<64; i++) {
        int k=(i+offset) & 63;
        // normal brightness
        set_ega_color(64+i, rgb[k][0], rgb[k][1], rgb[k][2]);
        // darkened brightness
        set_ega_color(192+i, rgb[k][0]>>2, rgb[k][1]>>2, rgb[k][2]>>2);
    }
}

// load a color scheme into the background color palette array by interpolation
void loadpal(int mode)
{
    struct ColorScheme *scheme = &schemes[mode%(sizeof(schemes)/sizeof(schemes[0]))];
    for (int kk=0; kk<32; kk++) {
#define BGMIX(comp) (((scheme->bg1.comp >> 2) * kk + \
                      (scheme->bg2.comp >> 4) * (31-kk))/31)
        rgb[63-kk][0] = rgb[kk][0] = BGMIX(r);
        rgb[63-kk][1] = rgb[kk][1] = BGMIX(g);
        rgb[63-kk][2] = rgb[kk][2] = BGMIX(b);
#undef BGMIX
    }
}

// current color schem
int g_scheme_idx = 1;
// current color scheme used for the title screen
int g_title_scheme_idx = 13;

// configure the color scheme for tiles at a given palette offset
// 6 entries are used:
//  - primary color
//  - secondary color
//  - lighter version of primary color
//  - lighter versino of secondary color
//  - darker version of primary color
//  - darker version of secondary color
void load_scheme(int offset, struct ColorScheme *scheme, int darken=100)
{
    int pr = darken * (scheme->primary.r >> 2) / 100;
    int pg = darken * (scheme->primary.g >> 2) / 100;
    int pb = darken * (scheme->primary.b >> 2) / 100;

    int sr = darken * (scheme->secondary.r >> 2) / 100;
    int sg = darken * (scheme->secondary.g >> 2) / 100;
    int sb = darken * (scheme->secondary.b >> 2) / 100;

    set_ega_color(offset, pr, pg, pb);
    set_ega_color(offset+1, sr, sg, sb);

#define LIGHTER(x) darken * (int)(63-(4*(63-x))/10) / 100
#define DARKER(x)  darken * (int)((6*x)/10) / 100

    set_ega_color(offset+2, LIGHTER(pr), LIGHTER(pg), LIGHTER(pb));
    set_ega_color(offset+3, LIGHTER(sr), LIGHTER(sg), LIGHTER(sb));

    set_ega_color(offset+4, DARKER(pr), DARKER(pg), DARKER(pb));
    set_ega_color(offset+5, DARKER(sr), DARKER(sg), DARKER(sb));
}

// Switch to the next tileset (callback from gamelogic.h)
void change_tileset()
{
    tileset = (tileset + 1) % (sizeof(packed_tiles)/sizeof(packed_tiles[0]));
}

// Switch to the next color scheme (callback from gamelogic.h)
void change_colors()
{
    g_scheme_idx++;
    load_scheme(1, &schemes[g_scheme_idx%(sizeof(schemes)/sizeof(schemes[0]))]);
    loadpal(g_scheme_idx);
}

// Switch to the next background pattern (callback from gamelogic.h)
void change_background()
{
    diamondmode = (diamondmode + 1) % MAXDIAMOND;
}

// Configure our 256-color VGA palette fully for a given color scheme
void initialize_palette(int scheme)
{
    // black
    set_ega_color(0, 0, 0, 0);
    // tile colors + variants (index 1..6)    -- these are used for normal tile rendering
    load_scheme(1, &schemes[g_scheme_idx]);
    // 7 = reserved/unused
    // dark gray
    set_ega_color(8, 4, 4, 4);
    // white
    set_ega_color(9, 63, 63, 63);
    // text colors
    set_ega_color(10, 56, 56, 56);
    set_ega_color(11, 46, 46, 46);
    // text shadow color for "how to play" page
    set_ega_color(12, 20, 20, 20);
    // 13, 14, 15 = reserved/unused

    // 16, .. 32 = graphics palette (for the "bomb animation" 16-color graphic)
    for (int p=0; p<16; ++p) {
        set_ega_color(16+p, GRAPHICS_PALETTE[p].r>>2, GRAPHICS_PALETTE[p].g>>2, GRAPHICS_PALETTE[p].b>>2);
    }
    // 33, .. 63 = reserved/unused

    // 64 .. 127 = rotating palette 1 (normal) -- for background pattern
    // 128 .. 187 = first color scheme, 10 shades
    // (the shades are used in the intro screen for the perspective fade-to-black effect)
    // this ignores the "scheme" parameter and uses g_title_scheme_idx instead
    int darken = 100;
    for (int i=128; i<128+6*10; i+=6) {
        load_scheme(i, &schemes[g_title_scheme_idx], darken);
        darken -= 10;
    }
    // 188 .. 191 = reserved/unused
    // 192 .. 255 = rotating palette 2 (shaded) -- for background pattern with "semi-transparent black" on top

    // note that the rotating palette 2 (starts at 192) has an
    // offset of 128 from rotating palette 1 (starts at 64), so
    // that we can use the "orvalue" of 128 when blitting from
    // the offscreen diamond (background pattern) data to the
    // screen to render it as "semi-transparent black" (the
    // orvalue can be added by the VGA latches "for free").

    // this actually loads the RGB gradient + loads the 64..127 + 192..255 ranges
    loadpal(scheme);
    vgapalette(0);
}

// poor man's pre-calculated sine tab, used for the "circular" background pattern
static unsigned char sinetab[16] = {
    32, 43, 53, 60, 63, 60, 53, 43, 32, 20, 10, 3, 1, 3, 10, 20,
};

// Y offset lookup when shaking the screen (bomb explosion)
static unsigned char shaking[] = {
    10, 0, 5, 0, 3, 0,
};

// draw the loading screen progress bar
void load_progress(bool reset=false)
{
    static int x = 0;
    int w = 8;
    int h = 8;
    int max_load = 25;
    static int cur_load = 0;
    int init_x = (PORT_WIDTH - w*max_load) / 2;
    int y = (PORT_HEIGHT - h) / 2 + 2*8;

    if (reset || x == 0) {
        x = init_x;
        cur_load = 0;
    }

    if (cur_load >= max_load) {
        // Need to update max load here
        ++cur_load;
        return;
    }

    render_rect(x, y, w, h, 1);
    if (x == init_x) {
        render_rect(x, y, 1, h, 3);
    }
    render_rect(x+w, y, 1, h, 5);
    render_rect(x, y, w, 1, 3);
    render_rect(x, y+h-1, w, 1, 5);

    // TODO: Eventually call vgawait every now and then
    //vgawait();

    x += w;
    ++cur_load;
}

// poor man's "blur" effect that works on a rect in VGA memory
void blur(int x, int y, int w, int h, int progress=0)
{
    for (int yy=y; yy<y+h-1; yy++) {
        for (int xx=x; xx<x+w-1; xx++) {
            unsigned short v = (getpixel(xx, yy) +
                    getpixel(xx+1, yy) +
                    getpixel(xx, yy+1) +
                    getpixel(xx+1, yy+1)) >> 2;
            putpixel(xx, yy, v);
            putpixel(xx+1, yy, v);
            putpixel(xx, yy+1, v);
            putpixel(xx+1, yy+1, v);
        }

        if (progress && yy%30 == 0) {
            load_progress();
        }
    }
}

// pre-render background patterns into offscreen memory for fast blitting later
void prepare_diamond()
{
    int offset = 64;
    int y, x;

    // xor grid
    for (y=0; y<64; ++y) {
        for (x=0; x<64; ++x) {
            putpixel(x, y+OFFSCREEN_Y, offset+ (((x+32) ^ (y+16)) & 63));
        }
    }
    load_progress();
    blur(0, OFFSCREEN_Y, 64, 64);
    load_progress();

    for (y=0; y<64; ++y) {
        for (x=0; x<64; ++x) {
            putpixel(x+64+64, y+OFFSCREEN_Y, offset+ (x*y*y*x & 63));
            int x3o = x + 16;
            int y3o = y + 16;
            putpixel(x+64+64+64, y+OFFSCREEN_Y, offset+ ((x3o*x3o+y3o*y3o+x3o*y3o) & 63));
            x3o = x*3 + 16;
            y3o = y*3 + 16;
            putpixel(x+64+64+64+64, y+OFFSCREEN_Y, offset+ ((x3o*x3o+y3o*y3o+x3o*y3o) & 63));
        }
    }
    load_progress();

    int h2 = PORT_HEIGHT / 2;
    int w2 = PORT_WIDTH / 2;

    // circular / looney toons
    for (y=0; y<h2; ++y) {
        for (x=0; x<w2; ++x) {
            long acc = (x - w2);
            acc *= acc;
            acc += (y-h2) * (y-h2);
            acc >>= 10;

            putpixel(x,              64+OFFSCREEN_Y+y, offset+ sinetab[(63-acc)%16]);
            putpixel(PORT_WIDTH-1-x, 64+OFFSCREEN_Y-1+PORT_HEIGHT-y, offset+ sinetab[(63-acc)%16]);
            putpixel(x,              64+OFFSCREEN_Y-1+PORT_HEIGHT-y, offset+ sinetab[(63-acc)%16]);
            putpixel(PORT_WIDTH-1-x, 64+OFFSCREEN_Y+y, offset+ sinetab[(63-acc)%16]);
        }

        if (y%30==0) {
            load_progress();
        }
    }
    load_progress();
    blur(0, OFFSCREEN_Y+64, PORT_WIDTH, PORT_HEIGHT, 1);
    load_progress();

    // diagonal
    for (y=0; y<64; ++y) {
        for (x=0; x<64; ++x) {
            putpixel(x+64, y+OFFSCREEN_Y, offset+ ((x + y) & 56));
        }
    }
    load_progress();
}

// blit from offscreen VGA memory to somewhere else, tiling and OR-ing values as needed
void blit_tiled(int x, int y, int w, int h, int sx, int sy, int tw, int th, int orvalue)
{
    int x0 = x;
    //int y0 = y;
    int x1 = x + w;
    int y1 = y + h;

    int yoff = y % th;
    //int hh = h;

    while (y < y1) {
        x = x0;
        int xoff = x % tw;
        int hh = (y1 - y);
        if (hh > th - yoff) {
            hh = th - yoff;
        }
        while (x < x1) {
            int ww = (x1 - x);
            if (ww > tw - xoff) {
                ww = tw - xoff;
            }
            vgablit(sx+xoff, sy+yoff, x, y, ww, hh, orvalue);
            x += ww;
            xoff = 0;
        }
        y += hh;
        yoff = 0;
    }
}

// render background pattern to screen, optionally darkening (via palette index offset) the result
void diamond(int x, int y, int w, int h, int mode, int dark)
{
    // 128 is the offset between rotating palette 1 (starts at 64) and rotating palette 2 (starts at 192)
    dark = dark ? 128 : 0;
    switch (mode) {
        default:
        case 0: blit_tiled(x, y, w, h, 0, OFFSCREEN_Y, 64, 64, dark); break;
        case 1: blit_tiled(x, y, w, h, 0, OFFSCREEN_Y+64, PORT_WIDTH, PORT_HEIGHT, dark); break;
        case 2: blit_tiled(x, y, w, h, 64, OFFSCREEN_Y, 64, 64, dark); break;
        case 3: blit_tiled(x, y, w, h, 128, OFFSCREEN_Y, 64, 64, dark); break;
        case 4: blit_tiled(x, y, w, h, 192, OFFSCREEN_Y, 64, 64, dark); break;
        case 5: blit_tiled(x, y, w, h, 256, OFFSCREEN_Y, 64, 64, dark); break;
        break;
    }
}

// helper function for prepare_cells() below
void prepare_cell(int x0, int y0, int color)
{
    unsigned char tc = 1+(color - 1)%2;

    unsigned char mapping[4];
    mapping[0] = 0;
    mapping[1] = tc;
    mapping[2] = tc + 2;
    mapping[3] = tc + 4;
    if (color == 3 || color == 4) {
        // ready-for-queue (remove highlights)
        mapping[2] = mapping[3] = mapping[1];
    } else if (color == 5 || color == 6) {
        // queued (inverse alpha color)
        mapping[0] = tc;
        mapping[1] = mapping[2] = mapping[3] = tc + 2;
    }
    const unsigned char *src = tile_pixels;
    for (int yy=0; yy<game_layout.block.h; yy++) {
        for (int xx=0; xx<game_layout.block.w; xx++) {
            unsigned char v = mapping[*src++];
            putpixel(x0+xx, y0+yy, v);
        }
    }
}

// render pre-shaded copies of the current tile to offscreen VGA memory for fast blitting
void
prepare_cells()
{
    for (int ts=0; ts<sizeof(packed_tiles)/sizeof(packed_tiles[0]); ++ts) {
        load_tiles(ts);
        for (int j=1; j<=6; ++j) {
            prepare_cell(game_layout.block.w*(j-1), OFFSCREEN_Y-game_layout.block.h*(1+ts), j);
        }
    }
}

// render the bomb animation frames into offscreen memory for fast blitting
void
prepare_graphics()
{
    int x0 = 0;
    int y0 = OFFSCREEN_Y - game_layout.block.h*(1 + (sizeof(packed_tiles)/sizeof(packed_tiles[0])));

    for (int y=0; y<graphics.height; ++y) {
        for (int x=0; x<graphics.width; ++x) {
            uint8_t value = graphics.data[(y*graphics.width + x)/2];
            if (x%2 == 0) {
                value >>= 4;
            } else {
                value &= 0xF;
            }
            putpixel(x0 + x, y0 + y, (value != 0) ? (value + 16) : 0);
        }
    }
}

// blit the bomb animation into on-screen VGA memory
void
render_bomb_animation(int x0, int y0, int num)
{
    int xoff = 16*num;
    int yoff = OFFSCREEN_Y - game_layout.block.h*(1 + (sizeof(packed_tiles)/sizeof(packed_tiles[0])));

#if defined(__BORLANDC__)
    vgablit(xoff, yoff, x0, y0, 16, 16, 0x00);
#else
    diamond(x0, y0, 16, 16, diamondmode, 1);
    transparentblit(xoff, yoff, x0, y0, 16, 16);
#endif
}

// blit a block/tile into on-screen VGA memory
void render_cell(int x, int y, int color)
{
    int x0 = game_layout.play_field.x + x*game_layout.block.w;
    int y0 = game_layout.play_field.y + y*game_layout.block.h;


    if (color >= 1 && color <= 6) {
        int xoff = game_layout.block.w*(color-1);
        int yoff = OFFSCREEN_Y - game_layout.block.h*(1+tileset);
#if defined(__BORLANDC__)
        // Fast blitting, not having any transparency
        vgablit(xoff, yoff, x0, y0, game_layout.block.w, game_layout.block.h, 0x00);
#else
        // On non-DOS platforms, we're double-buffered, and we can waste some
        // cycles, so we can do the draw-diamond-then-overdraw-with-opaque-pixels
        // technique. TODO: Make this fast and working for DOS by composing in an
        // offscreen area, then blitting fast to the screen.
        diamond(x0, y0, game_layout.block.w, game_layout.block.h, diamondmode, 1);
        transparentblit(xoff, yoff, x0, y0, game_layout.block.w, game_layout.block.h);
#endif
    } else if (color >= 7) {
        render_bomb_animation(x0, y0, color - 7);
    } else {
        diamond(x0, y0, game_layout.block.w, game_layout.block.h, diamondmode, 1);
    }
}

enum DrawCorners {
    DRAW_CORNER_TOPLEFT = 1 << 0,
    DRAW_CORNER_BOTTOMLEFT = 1 << 1,
    DRAW_CORNER_TOPRIGHT = 1 << 2,
    DRAW_CORNER_BOTTOMRIGHT = 1 << 3,
    DRAW_CORNER_TOP = DRAW_CORNER_TOPLEFT | DRAW_CORNER_TOPRIGHT,
    DRAW_CORNER_BOTTOM = DRAW_CORNER_BOTTOMLEFT | DRAW_CORNER_BOTTOMRIGHT,
    DRAW_CORNER_LEFT = DRAW_CORNER_TOPLEFT | DRAW_CORNER_BOTTOMLEFT,
    DRAW_CORNER_RIGHT = DRAW_CORNER_TOPRIGHT | DRAW_CORNER_BOTTOMRIGHT,
    DRAW_CORNER_ALL = DRAW_CORNER_TOP | DRAW_CORNER_BOTTOM,
};

// corners of a rounded rectangle
void
draw_corners_rect(int x0, int y0, int w, int h, int bg_outer=9, int bg_inner=11, int bg_border=9, int which=DRAW_CORNER_ALL)
{
    static unsigned char corner[6*6] = {
        1, 1, 1, 1, 1, 2,
        1, 1, 1, 2, 0, 0,
        1, 1, 2, 0, 0, 0,
        1, 2, 0, 0, 0, 0,
        1, 0, 0, 0, 0, 0,
        2, 0, 0, 0, 0, 0,
    };

    int m[3] = {
        bg_inner,
        bg_outer,
        bg_border,
    };

    for (int i=0; i<36; i++) {
        int y = i/6;
        int x = i%6;
        int color = m[corner[i]];
        if (color == -1) { continue; }
        if (which & DRAW_CORNER_TOPLEFT) {
            putpixel(x0+x,     y0+y, color);
        }
        if (which & DRAW_CORNER_BOTTOMLEFT) {
            putpixel(x0+x,     y0+h-1-y, color);
        }
        if (which & DRAW_CORNER_TOPRIGHT) {
            putpixel(x0+w-1-x, y0+y, color);
        }
        if (which & DRAW_CORNER_BOTTOMRIGHT) {
            putpixel(x0+w-1-x, y0+h-1-y, color);
        }
    }
}

// convenience function for screen corners
static void
draw_corners(int bg_outer=9, int bg_inner=11, int inset=0)
{
    draw_corners_rect(inset, inset, PORT_WIDTH-2*inset, PORT_HEIGHT-2*inset, bg_outer, bg_inner, bg_outer);
}

struct ScoreDisplay {
    LayoutRect *rect;
    const char **label;
    score_t *value;
    score_t last_value;
    unsigned char color;
};

const int
ScoreDisplay_inset = 6;

// re-draw the background layer of the screen + the scores
static void
redraw_bg(ScoreDisplay *scores)
{
    int i;
    for (i=0; i<sizeof(subdivided)/sizeof(subdivided[0]); ++i) {
        if (subdivided[i].kind == 0) {
            // dark
            diamond(subdivided[i].x, subdivided[i].y, subdivided[i].w, subdivided[i].h, diamondmode, 1);
        } else if (subdivided[i].kind == 1) {
            // light
            diamond(subdivided[i].x, subdivided[i].y, subdivided[i].w, subdivided[i].h, diamondmode, 0);
        }
    }

    int spacing = 4;
    int height = 8 + spacing + 8;
    int line1_width = 8 * mystrlen(APPNAME);
    int line2_width = 8 * mystrlen(REVISION_STRING);
    if (game_layout.title_bar.h >= height) {
        for (i=0; i<2; i++) {
            render_font(game_layout.title_bar.x + (game_layout.title_bar.w - line1_width) / 2, game_layout.title_bar.y + (game_layout.title_bar.h - height) / 2, APPNAME, 9*i);
            render_font(game_layout.title_bar.x + (game_layout.title_bar.w - line2_width) / 2, game_layout.title_bar.y + (game_layout.title_bar.h - height) / 2 + 8 + spacing, REVISION_STRING, 11*i);
        }
    } else {
        height = 8;
        for (i=0; i<2; i++) {
            render_font(game_layout.title_bar.x + spacing, game_layout.title_bar.y + (game_layout.title_bar.h - height) / 2, APPNAME, 9*i);
            render_font(game_layout.title_bar.x + spacing + line1_width + spacing, game_layout.title_bar.y + (game_layout.title_bar.h - height) / 2, REVISION_STRING, 11*i);
        }
    }

    while (scores->label) {
        diamond(scores->rect->x, scores->rect->y, scores->rect->w, scores->rect->h, diamondmode, 1);
        int textwidth = (scores->rect->w - 2*ScoreDisplay_inset) / 8;
        render_font_wrap(scores->rect->x+ScoreDisplay_inset+1, scores->rect->y+ScoreDisplay_inset, *(scores->label), 0, textwidth);
        render_font_wrap(scores->rect->x+ScoreDisplay_inset, scores->rect->y+ScoreDisplay_inset, *(scores->label), 9, textwidth);

        // Force redraw of scores
        scores->last_value = 0xFFFFFFFFUL;

        ++scores;
    }

    draw_corners(0, -1);
}

void
render_button(int x, int y, int w, int h, const char *text, int fg, int bg, int bd, int outer_color=-1, int which=DRAW_CORNER_ALL, int active=0)
{
    // top
    render_rect(x+6,   y,      w-2*6, 6, bg);
    // bottom
    render_rect(x+6,   y+h-6,  w-2*6, 6, bg);
    // left
    render_rect(x,     y+6,    6,     h-2*6, bg);
    // right
    render_rect(x+w-6, y+6,    6,     h-2*6, bg);
    // inner
    render_rect(x+6,   y+6,    w-2*6, h-2*6, bg);

    // corners
    draw_corners_rect(x, y, w, h, outer_color, bg, bd, which);

    if ((which & DRAW_CORNER_TOPLEFT) == 0) {
        render_rect(x, y, 6, 6, bg);
    }
    if ((which & DRAW_CORNER_TOPRIGHT) == 0) {
        render_rect(x+w-6, y, 6, 6, bg);
    }
    if ((which & DRAW_CORNER_BOTTOMLEFT) == 0) {
        render_rect(x, y+h-6, 6, 6, bg);
    }
    if ((which & DRAW_CORNER_BOTTOMRIGHT) == 0) {
        render_rect(x+w-6, y+h-6, 6, 6, bg);
    }

    int max_chars = (w - 12) / 8;

    if (max_chars > mystrlen(text)) {
        max_chars = mystrlen(text);
    }

    if (mystrlen(text) > max_chars) {
        int scroll_time_ms_per_char = 500;
        int xpixels = (mystrlen(text) - max_chars);
        int off = (time_now_ms() / scroll_time_ms_per_char) % (2 * xpixels + 2);
        int xoff = 4 - 8 * (time_now_ms() % scroll_time_ms_per_char) / scroll_time_ms_per_char;

        if (off > xpixels) {
            off = 2*xpixels + 1 - off;
            xoff *= -1;
        }

        if (!active) {
            off = xoff = 0;
        }

        const char *tmp = text + off;

        int x0 = x + (w - 8 * max_chars) / 2;
        render_font_xclipped(x0 + xoff, y + (h - 8) / 2, tmp, fg, x0, x0+8*max_chars-xoff, max_chars);
    } else {
        render_font(x + (w - 8*max_chars)/2, y + (h - 8) / 2, text, fg);
    }
}

// convenience function using the current color scheme
void
render_button_themed(int x, int y, int w, int h, const char *text, int active, int which=DRAW_CORNER_ALL, int outer_color=-1)
{
    int inactive_bg = 128;
    int inactive_fg = 128 + 4;
    int inactive_bd = 128 + 4;

    int active_bg = 128 + 2;
    int active_fg = 128 + 4;
    int active_bd = 128;

    int fg = active ? active_fg : inactive_fg;
    int bg = active ? active_bg : inactive_bg;
    int bd = active ? active_bd : inactive_bd;

    render_button(x, y, w, h, text, fg, bg, bd, outer_color, which, active);
}

// called whenever the language changes in the options menu
void update_language()
{
    lstrings_choose(hw_cfg.language);
}

// convenience class for dealing with double-buffering
struct DoubleBuffering {
    DoubleBuffering()
        : backbuffer_y(PORT_HEIGHT)
        , frontbuffer_y(0)
        , cleared(0)
    {
        // display front buffer
        vgascroll(frontbuffer_y);
        vgawait();
        // clear back buffer
        clear();
    }

    ~DoubleBuffering()
    {
        //clear();
        //backbuffer_y = frontbuffer_y;
        //clear();
        vgascroll(0);
    }

    int y() { return backbuffer_y; }
    int front_y() { return frontbuffer_y; }

    void swap() {
        int tmp = backbuffer_y;
        backbuffer_y = frontbuffer_y;
        frontbuffer_y = tmp;
        vgascroll(frontbuffer_y);
        if (cleared == 0) {
            clear();
            cleared = 1;
        }
    }

    void clear() {
        vgaclearbytes_y(backbuffer_y, (long)PORT_WIDTH*(long)PORT_HEIGHT);
    }

    int backbuffer_y;
    int frontbuffer_y;
    int cleared;
};

struct SettingsTab {
    // indirect pointer, as the value might change at runtime when changing language
    const char **label;
    unsigned char *dest;
    unsigned char max;
    const char **value_labels;
    void (*update)();
};

bool hwsetup_page(const char **title, SettingsTab *tabs, int n)
{
    int grid = PORT_HEIGHT / 10;

    // TODO: Maybe add this to the layout?
    int padding = 4;
    int max_width = (PORT_WIDTH - 2 * grid - 20 - padding * (n-1)) / n;

    int y0 = 2 * grid;
    int x0 = grid;
    int button_width = max_width;
    int button_height = padding + 8 + padding;
    int offset = 2;

    int active = 0;
    int do_redraw_bg = 2;

    int area_width = PORT_WIDTH - x0 - grid;
    int area_height = PORT_HEIGHT - y0 - grid;

    Timestep pal(10);

    time_ms_t last = time_ms_t();

    DoubleBuffering dblbuf;

    while (true) {
        y0 = dblbuf.y() + 2 * grid;

        int xpos = (PORT_WIDTH - 2 * 8 * mystrlen(*title)) / 2;
        int ypos = dblbuf.y() + (2 * grid - 2 * 8) / 2;

        if (do_redraw_bg) {
            // clear previous title choice string
            render_rect(0, ypos, PORT_WIDTH, 2*8, 0);

            // render current title choice string
            render_font_x(xpos-1, ypos, *title, 79, 2);
            render_font_x(xpos+1, ypos, *title, 69, 2);
            render_font_x(xpos, ypos, *title, 9, 2);

            render_rect(x0, y0 + button_height, area_width, area_height, 128 + 2);
            draw_corners_rect(x0, y0 + button_height, area_width, area_height, 0, 128 + 2, 128 + 4, DRAW_CORNER_BOTTOM | DRAW_CORNER_TOPRIGHT);
            --do_redraw_bg;
        }

        int i;

        if (n != 1) {
            for (i=0; i<n; ++i) {
                if (i != active) {
                    render_rect(x0 + i * (button_width + padding), y0 - offset, button_width, offset, 0);
                }
                render_button_themed(x0 + i * (button_width + padding), y0 - offset * (i == active), button_width, button_height + offset * (i == active), *(tabs[i].label), (i == active), DRAW_CORNER_TOP, 0);
            }

            if (active == 0) {
                draw_corners_rect(x0, y0 + button_height, area_width, area_height, 128 + 2, 128 + 2, 128 + 2, DRAW_CORNER_TOPLEFT);
            } else {
                draw_corners_rect(x0, y0 + button_height, area_width, area_height, 0, 128 + 2, 128 + 4, DRAW_CORNER_ALL);
                draw_corners_rect(x0, y0 + button_height, area_width, area_height, 128, 128 + 2, 128 + 2, DRAW_CORNER_TOPLEFT);
            }
        } else {
            draw_corners_rect(x0, y0 + button_height, area_width, area_height, 0, 128 + 2, 128 + 4, DRAW_CORNER_ALL);
        }

        SettingsTab &tab = tabs[active];

        int x1 = x0 + padding * 3;
        int y1 = y0 + button_height + padding * 3;

        for (i=0; i<tab.max; ++i) {
            int off = 0;
            if (*(tab.dest) == i) {
                off = 3;
                render_font(x1 + off, y1 + i * 10, ">", 128);
                render_font(x1 + off + 8 + 8 * mystrlen(tab.value_labels[i]), y1 + i * 10, "<", 128);
            }
            render_font(x1 + off + 8, y1 + i * 10, tab.value_labels[i], 128 + 4);
        }

        PortEvent e;
        if (event_poll(&e)) {
            if (e.type == EVENT_KEYDOWN) {
                if (is_enter(e.key)) {
                    vgascroll(0);
                    return true;
                } else if (is_back(e.key)) {
                    vgascroll(0);
                    return false;
                } else if (is_right(e.key)) {
                    if (active < n-1) {
                        active++;
                        do_redraw_bg = 2;
                    }
                } else if (is_left(e.key)) {
                    if (active > 0) {
                        active--;
                        do_redraw_bg = 2;
                    }
                } else if (is_up(e.key)) {
                    if (*(tab.dest) > 0) {
                        (*(tab.dest))--;
                        if (tab.update) {
                            tab.update();
                        }
                        do_redraw_bg = 2;
                    }
                } else if (is_down(e.key)) {
                    if (*(tab.dest) < tab.max-1) {
                        (*(tab.dest))++;
                        if (tab.update) {
                            tab.update();
                        }
                        do_redraw_bg = 2;
                    }
                }
            }
        }

        time_ms_t now = time_now_ms();
        time_ms_t add = now - last;
        last = now;

        pal.update(add);

        while (pal.step()) {
            palrotate++;
        }

        vgapalette(palrotate);

        dblbuf.swap();
        vgawait();
    }

    return false;
}

// options screen main loop
void hwsetup()
{
    const char *LEVEL_STRINGS[] = {"1", "2", "3", "4", "5", "6", "7", "8", "9", "10"};

    SettingsTab tabs[] = {
        { &HWCONFIG_LANGUAGE_TITLE, &hw_cfg.language, LANGUAGE_MAX, LANGUAGES, update_language },
#if !defined(AUTOCONFIGURE_AUDIO_DEVICE)
        { &HWCONFIG_AUDIO_DEVICE_TITLE, &hw_cfg.audio_dev, AUDIO_DEV_MAX, AUDIO_DEVICES, 0 },
#endif
        { &HWCONFIG_SOUND_TRACK_TITLE, &hw_cfg.sound_track, SOUND_TRACK_MAX, SOUND_TRACKS, 0 },
        { &HWCONFIG_STARTING_LEVEL_TITLE, &hw_cfg.starting_level, sizeof(LEVEL_STRINGS)/sizeof(LEVEL_STRINGS[0]), LEVEL_STRINGS, 0},
    };

#if defined(__BORLANDC__)
    int prev_audio_dev = hw_cfg.audio_dev;

again:
#endif
    if (!hwsetup_page(&TITLE_CHOICE_STR_OPTIONS, tabs, sizeof(tabs)/sizeof(tabs[0]))) {
        // User exited the options screen
        return;
    }

#if defined(__BORLANDC__)
    if (prev_audio_dev != hw_cfg.audio_dev) {
        if (hw_cfg.audio_dev == AUDIO_DEV_OPL2) {
            SettingsTab tabs[] = {
                { &(AUDIO_DEVICES[2]), &hw_cfg.opl_setup, AUDIO_OPL_MAX, AUDIO_OPTIONS_OPL, 0 },
            };
            if (!hwsetup_page(&HWCONFIG_AUDIO_DEVICE_TITLE, tabs, 1)) {
                goto again;
            }
        }

        if (hw_cfg.audio_dev == AUDIO_DEV_MPU401) {
            SettingsTab tabs[] = {
                { &(AUDIO_DEVICES[3]), &hw_cfg.mpu_setup, AUDIO_MPU_MAX, AUDIO_OPTIONS_MPU, 0 },
            };
            if (!hwsetup_page(&HWCONFIG_AUDIO_DEVICE_TITLE, tabs, 1)) {
                goto again;
            }
        }
    }
#endif
}

struct OffsetInfo {
    unsigned char mapping[4], *tilebase, *vfieldline;
};

struct LeftRight {
    unsigned char x0, x1;
};

struct Choice {
    // indirect pointer, as the value might change at runtime when language changes
    const char **label_ptr;
    int value;
};

void
render_vertical_buttons(struct Choice *choices, int n, int active, int y0, int y1, int only=-1)
{
    // TODO: Maybe add this to the layout?
    int spacing = (PORT_HEIGHT == 240) ? 8 : 12; // 12 on PSP, 8 on 3DS
    int padding = (PORT_HEIGHT == 240) ? 4 : 8; // 8 on PSP, 4 on 3DS

    int i;

    int max_width = 120;

    int button_height = (2 * padding + 8);
    int total_height = n * button_height + (n - 1) * spacing;
    int button_offset = button_height + spacing;
    int total_width = 2 * padding + max_width;

    int x = (PORT_WIDTH - total_width) / 2;
    int y = y0 + ((y1 - y0) - total_height) / 2;

    for (i=0; i<n; ++i) {
        if (only == -1 || only == i) {
            render_button_themed(x, y + i * button_offset, total_width, button_height, *(choices[i].label_ptr), (i == active));
        }
    }
}

enum TitlescreenResult {
    TITLESCREEN_PLAY,
    TITLESCREEN_OPTIONS,
    TITLESCREEN_HOW_TO_PLAY,
    TITLESCREEN_QUIT,
};

struct Choice choices[] = {
    { &TITLE_CHOICE_STR_PLAY, TITLESCREEN_PLAY },
    { &TITLE_CHOICE_STR_OPTIONS, TITLESCREEN_OPTIONS },
    { &TITLE_CHOICE_STR_HOW_TO_PLAY, TITLESCREEN_HOW_TO_PLAY },
    { &TITLE_CHOICE_STR_QUIT, TITLESCREEN_QUIT },
};

struct TextScreenContext {
    int linespacing;
    int height;
    int x;
    int y;
    int cur;
};

void text_screen_render_line(const char *line, int n, void *user_data)
{
    struct TextScreenContext *ctx = (struct TextScreenContext *)user_data;

    int y = ctx->y + (ctx->linespacing + 8)*ctx->cur;

    render_font(ctx->x-1, y, line, (ctx->cur == 0) ? 5 : ((ctx->cur == 1) ? 6 : 12), n);
    render_font(ctx->x,   y, line, (ctx->cur == 0) ? 1 : ((ctx->cur == 1) ? 2 : 9), n);

    ctx->cur++;
}

// render a text screen (first render offscreen, then swap)
void textscreen(const char **lines)
{
    DoubleBuffering dblbuf;

    int n = 0;
    int border = 10;
    int max_chars_per_line = (PORT_WIDTH - 2*border) / 8;
    if (max_chars_per_line > 40) {
        max_chars_per_line = 40;
    }
    int j = 0;
    while (lines[j]) {
        n += line_split(lines[j++], max_chars_per_line, 0, 0);
    }

    struct TextScreenContext ctx;
    ctx.linespacing = 1;
    ctx.height = (n - 1) * ctx.linespacing + n * 8;
    ctx.x = (PORT_WIDTH - 8*max_chars_per_line) / 2;
    ctx.y = dblbuf.y() + (PORT_HEIGHT - ctx.height) / 2;
    ctx.cur = 0;

    for (int kk=0; lines[kk]; ++kk) {
        line_split(lines[kk], max_chars_per_line, text_screen_render_line, &ctx);
    }

    dblbuf.swap();

    while (true) {
        vgawait();

        PortEvent e;
        if (event_poll(&e)) {
            if (e.type == EVENT_KEYDOWN && e.key == KeyCode_Space) {
                break;
            }
        }
    }
}

int active_choice = 0;

// render the title screen menu (double-buffered)
int titlescreen(MusicPlayer *music)
{
    DoubleBuffering dblbuf;

    if (hw_cfg.audio_dev == AUDIO_DEV_CDROM) {
        cd_stop();
        cd_play_track(4);
    }

    int local_tileset = 0;
    load_tiles(local_tileset);

    time_ms_t last = time_now_ms();
    Timestep fall(70);
    Timestep tiles(7000);
    Timestep pal(10);
    Timestep beat(140);

#define VFIELDY 4
    unsigned char vfield[8*VFIELDY];

    int vfieldoffset = 0;
    int last_vfieldoffset = -1;
    for (int i=0; i<8*VFIELDY; i++) {
        vfield[i] = (i+(i/7)+(i/11))%3;
    }

    int draw_bg = 2;

    int index = 0;

    static unsigned char yoff[] = {
        60,
        62,  // 6
        64,
        66,
        70,
        74,
        80,
        86,
        92,  // 8
        100,
        108,
        116,
        124,
        132,
        140, // 10
        150,
        160,
        171,
    };

    int TITLE_X_BEGIN = 30;
    int TITLE_X_END = 150;

#define TITLE_Y_BEGIN 170
#define TITLE_Y_END 60
#define MAX_WIDTH 120

    struct OffsetInfo offsets[MAX_WIDTH+1];

    // size determined by printing rioff at the end of the init loop
    unsigned char rowinfo[8742];

    struct LeftRight lr[TITLE_Y_BEGIN-TITLE_Y_END+1];

    {
        int x0S = TITLE_X_BEGIN<<3;
        int x1S = TITLE_X_END<<3;

        int rioff = 0;
        struct LeftRight *llr = lr;
        for (int y=TITLE_Y_BEGIN; y>=TITLE_Y_END; --y) {
            llr->x0 = x0S >> 3;
            llr->x1 = x1S >> 3;

            int blckS = (llr->x1 - llr->x0);

            ++llr;

            for (int xx=0; xx<blckS; xx++) {
                int xxS = xx<<3;
                int xcoord = xxS/blckS;
                int xoffset = game_layout.block.w*(xxS%blckS)/blckS;
                rowinfo[rioff++] = (xcoord << 4) | xoffset;
            }

            x0S -= 1;
            x1S -= 7;
        }
    }

    while (true) {
        time_ms_t now = time_now_ms();
        time_ms_t add = now - last;
        last = now;

        if (draw_bg) {
            --draw_bg;
            int y = dblbuf.y();

            int inset = 5;
            int bgc = 64+47;
            // bg inset
            render_rect(0, y + 0, PORT_WIDTH, inset, bgc);
            render_rect(0, y + PORT_HEIGHT-inset, PORT_WIDTH, inset, bgc);
            render_rect(0, y + inset, inset, PORT_HEIGHT-2*inset, bgc);
            render_rect(PORT_WIDTH-inset, y + inset, inset, PORT_HEIGHT-2*inset, bgc);

            draw_corners_rect(inset, y + inset, PORT_WIDTH-2*inset, PORT_HEIGHT-2*inset, bgc, -1, bgc);

            int xpos = (PORT_WIDTH - 8 * 2 * mystrlen(APPNAME)) / 2;
            render_font_x(xpos-1, y+30, APPNAME, 79, 2);
            render_font_x(xpos+1, y+30, APPNAME, 69, 2);
            render_font_x(xpos, y+30, APPNAME, 9, 2);

            render_font((PORT_WIDTH - 8 * mystrlen(TAGLINE)) / 2, y+50, TAGLINE, 11);

#if defined(__BORLANDC__)
            const char *tech = TECHNOLOGIES_DOS;
#else
            const char *tech = TECHNOLOGIES;
#endif
            const char *tech1 = tech + mystrlen(tech) + 1;
            const char *tech2 = tech1 + mystrlen(tech1) + 1;
            int tech_x = (PORT_WIDTH - 8 * ((tech2 - tech) + mystrlen(tech2))) / 2;

            int tech_y = y + 60;
            render_font(tech_x-1, tech_y, tech, 157);
            render_font(tech_x, tech_y, tech, 153);
            render_font(tech_x+(tech1-tech)*8, tech_y, tech1, 10);
            render_font(tech_x+(tech2-tech)*8+1, tech_y, tech2, 156);
            render_font(tech_x+(tech2-tech)*8, tech_y, tech2, 128);

            int copy_x = (PORT_WIDTH - 8 * mystrlen(COPYRIGHT_LINE)) / 2;
            render_font(copy_x, y + PORT_HEIGHT - 22, COPYRIGHT_LINE, 11);
            render_font(PORT_WIDTH-10-8*mystrlen(REVISION_STRING), y + 10, REVISION_STRING, 74);
        }

        fall.update(add);
        tiles.update(add);
        pal.update(add);
        beat.update(add);

        while (fall.step()) {
            vfieldoffset++;
        }

        while (tiles.step()) {
            local_tileset = (local_tileset + 1) % (sizeof(packed_tiles)/sizeof(packed_tiles[0]));
            load_tiles(local_tileset);
        }

        while (pal.step()) {
            palrotate++;
        }

        while (beat.step()) {
            static int introbeat = 0;
            introbeat = (introbeat + 1) % 8;
            switch (introbeat) {
                case 0:
                case 2:
                case 3: music->fx(AFX_BASS_DRUM); break;
                case 5:
                case 6: music->fx(AFX_SNARE_DRUM); break;
                case 7: music->fx(AFX_HIHAT); break;
                default:
                    break;
            }
        }

        PortEvent e;
        while (event_poll(&e)) {
            if (e.type == EVENT_QUIT) {
                return TITLESCREEN_QUIT;
            } else if (e.type == EVENT_KEYDOWN) {
                if (is_back(e.key)) {
                    active_choice = sizeof(choices) / sizeof(choices[0]) - 1;
                    return choices[active_choice].value;
                } else if (is_enter(e.key)) {
                    time_ms_t started = time_now_ms();
                    while (true) {
                        time_ms_t now = time_now_ms();
                        time_ms_t elapsed = now - started;
                        if (elapsed > 400) {
                            break;
                        }
                        int i = elapsed / 100;
                        render_vertical_buttons(choices, sizeof(choices) / sizeof(choices[0]), (i%2) ? active_choice : -1, dblbuf.front_y() + PORT_HEIGHT/8, dblbuf.front_y() + PORT_HEIGHT, active_choice);
                        vgawait();
                    }
                    // Remove any key presses from the queue during the animation
                    while (event_poll(&e)) {}
                    return choices[active_choice].value;
                } else if (is_up(e.key)) {
                    if (active_choice > 0) {
                        active_choice--;
                    }
                } else if (is_down(e.key)) {
                    if (active_choice < sizeof(choices) / sizeof(choices[0]) - 1) {
                        active_choice++;
                    }
                } else if (is_left(e.key)) {
                    if (g_title_scheme_idx > 0) {
                        g_title_scheme_idx--;
                    } else {
                        g_title_scheme_idx = sizeof(schemes)/sizeof(schemes[0])-1;
                    }
                    initialize_palette(g_title_scheme_idx);
                } else if (is_right(e.key)) {
                    g_title_scheme_idx = (g_title_scheme_idx + 1) % (sizeof(schemes)/sizeof(schemes[0]));
                    initialize_palette(g_title_scheme_idx);
                }
            }
        }

        //time_ms_t begin_draw = time_now_ms();

        unsigned char scanline[MAX_WIDTH+1];

        if (last_vfieldoffset != vfieldoffset) {
            for (int y=TITLE_Y_END; y<=TITLE_Y_BEGIN; ++y) {
                int ycoord = 0;
                for (int i=0; i<sizeof(yoff)/sizeof(yoff[0]); ++i) {
                    if (y >= yoff[i]) {
                        ycoord++;
                    } else {
                        int yoffset = game_layout.block.h*(y-yoff[i-1]) / (yoff[i] - yoff[i-1]);

                        int basecolor = 128 + 6 * ((20-ycoord)/2);
                        offsets[y-TITLE_Y_END].mapping[0] = 0 + basecolor;
                        offsets[y-TITLE_Y_END].mapping[1] = 0 + basecolor;
                        offsets[y-TITLE_Y_END].mapping[2] = 2 + basecolor;
                        offsets[y-TITLE_Y_END].mapping[3] = 4 + basecolor;

                        offsets[y-TITLE_Y_END].tilebase = tile_pixels + yoffset * game_layout.block.h;
                        offsets[y-TITLE_Y_END].vfieldline = vfield + ((ycoord-(vfieldoffset%VFIELDY)+VFIELDY)%VFIELDY)*8;

                        break;
                    }
                }
            }

            last_vfieldoffset = vfieldoffset;
        }

        struct OffsetInfo *off = offsets + (TITLE_Y_BEGIN - TITLE_Y_END);

        const unsigned char *ri = rowinfo;
        struct LeftRight *llr = lr;
        for (int y=TITLE_Y_BEGIN; y>=TITLE_Y_END; --y) {
            int x0 = llr->x0;
            int x1 = llr->x1;
            ++llr;

            unsigned char *tilebase = off->tilebase;
            unsigned char *vfieldline = off->vfieldline;
            unsigned char *mapping = off->mapping;

            --off;

            int blckS = (x1 - x0);

            for (int xx=0; xx<blckS; xx++) {
                unsigned char tile = *ri++;
                unsigned char xcoord = (tile >> 4) & 0x0F;
                unsigned char item = vfieldline[xcoord];

                unsigned char color = 0;
                if (item) {
                    unsigned char xoffset = tile & 0x0F;
                    unsigned char v = tilebase[xoffset];
                    if (v) {
                        color = mapping[v] + (item-1);
                    }
                }

                scanline[xx] = color;
            }

            vgablitmemline(x0, dblbuf.y()+y+PORT_HEIGHT-200, scanline, blckS);
            vgablitmemline(PORT_WIDTH-x0-blckS, dblbuf.y()+y+PORT_HEIGHT-200, scanline, blckS);
        }

        //time_ms_t end_draw = time_now_ms();
        //render_rect(200, 10, 100, 10, 0);
        //render_number(300, 10, end_draw - begin_draw, 1);

        render_vertical_buttons(choices, sizeof(choices) / sizeof(choices[0]), active_choice, dblbuf.y() + PORT_HEIGHT/8, dblbuf.y() + PORT_HEIGHT);

        vgapalette(palrotate);
        ++index;

        dblbuf.swap();
        vgawait();
    }

    return TITLESCREEN_QUIT;
}

// render a yes/no question dialog and wait for input
// (palette cycling keeps going on)
static bool
ask_yes_no(const char *msg)
{
    const char *end = msg;
    while (*end) end++;
    int len = (end - msg);

    Timestep ts_palette(50UL);
    time_ms_t last = time_now_ms();

    int border = 12;
    int padding = 4;
    int spacing = 20;
    int button_spacing = 2 * padding;
    int question_width = 8 * len;
    int width = border + question_width + border;
    int height = border + 8 + spacing + padding + 8 + padding + border;
    int x = (PORT_WIDTH - width) / 2;
    int y = (PORT_HEIGHT - height) / 2;
    int bevel = 1;

    render_rect(x-bevel, y-bevel, width+2*bevel, height+2*bevel, 72);
    render_rect(x, y, width, height, 0);
    render_font(x + border, y + border, msg, 9);

    int button_width = (question_width - button_spacing) / 2;

    int old_selected = -1;
    int selected = 1;

    while (true) {
        time_ms_t now = time_now_ms();
        time_ms_t add = now - last;
        last = now;

        ts_palette.update(add);
        while (ts_palette.step()) palrotate++;

        PortEvent e;
        while (event_poll(&e)) {
            if (e.type == EVENT_QUIT) {
                return false;
            } else if (e.type == EVENT_KEYDOWN) {
                if (is_enter(e.key)) {
                    return (selected == 0);
                } else if (is_left(e.key)) {
                    selected = 0;
                } else if (is_right(e.key)) {
                    selected = 1;
                } else if (is_back(e.key)) {
                    return false;
                }
            }
        }

        if (selected != old_selected) {
            old_selected = selected;

            render_button_themed(x + border, y + border + 8 + spacing, button_width, padding + 8 + padding, YESNO_YES, (selected == 0));
            render_button_themed(x + border + button_width + button_spacing, y + border + 8 + spacing, button_width, padding + 8 + padding, YESNO_NO, (selected == 1));
        }

        vgawait();
        vgapalette(palrotate);
    }
}

MusicPlayer music;

// at exit, clean up and go back to text mode, remove interrupt handlers, etc..
void
loonies_cleanup()
{
    vgaclear();
    vgamode(VGAMode_Text);
    timer_quit();
    music.shutdown();
}

// callback is called once every millisecond, push the music playback forward
void on_timer()
{
    music.update(1);
}

// loading screen, load and unpack all assets / render background patterns offscreen
void load_assets()
{
    vgaclearbytes(4L*(long)PORT_WIDTH*(long)PORT_HEIGHT);
    render_font((PORT_WIDTH-8*mystrlen(LOADING))/2-1, (PORT_HEIGHT-8)/2, LOADING, 5);
    render_font((PORT_WIDTH-8*mystrlen(LOADING))/2, (PORT_HEIGHT-8)/2, LOADING, 1);
    vgawait();

    load_progress(true);
    load_digits();
    load_progress();

    prepare_diamond();

    // Loading completed
    vgaclear();
    vgawait();

    g_scheme_idx = 1;
    initialize_palette(g_scheme_idx);
}

int
loonies_main()
{
    // decode string data in memory
    lstrings_init();
    lstrings_choose(hw_cfg.language);

    // load savegame if it exists and hasn't been tampered with
    SaveGame save;
    if (!load_file(SAVE_FILENAME, &save, sizeof(save)) || !save.verify()) {
        save.set(0);
    }

    vgaclear();
    vgamode(VGAMode_320x240);

    // load color scheme before doing anything else
    g_scheme_idx = 2;
    initialize_palette(g_scheme_idx);

    // register timer for input handling + music
    timer_init();
    set_timer_callback(on_timer);

    // read out all existing events (clear event queue)
    PortEvent e;
    while (event_poll(&e)) {}

    // unpack asset data and render background patterns (loading screen)
    load_assets();

#if defined(AUTOCONFIGURE_AUDIO_DEVICE)
    hw_cfg.audio_dev = cd_available() ? AUDIO_DEV_CDROM : AUDIO_DEV_MPU401;
#endif

    // "outer main loop"
    while (true) {
        music.shutdown();
        music.configure(&hw_cfg);
        vgascroll(0);

        while (true) {
            switch (titlescreen(&music)) {
                case TITLESCREEN_PLAY:
                    goto start_game;
                    break;
                case TITLESCREEN_OPTIONS:
                    // stop current audio device
                    music.shutdown();
                    hwsetup();
                    // restart audio device with new config
                    music.configure(&hw_cfg);
                    break;
                case TITLESCREEN_HOW_TO_PLAY:
                    music.shutdown();
                    textscreen(INSTRUCTIONS);
                    music.configure(&hw_cfg);
                    break;
                case TITLESCREEN_QUIT:
                    if (ask_yes_no(QUESTION_REALLYQUIT)) {
                        loonies_cleanup();
                        return 1;
                    }
                default:
                    break;
            }
        }

    start_game:
        bool redraw = false;
        int last_diamondmode = -1;

        // clear 2 screens worth of VGA memory, the layout we use:
        //  - first screen is the visibile screen
        //  - second screen (offscreen) contains things to blit (tiles, digits, bomb, etc..)
        vgaclearbytes(2L*(long)PORT_WIDTH*(long)PORT_HEIGHT);

        // Wait to swap to initial screen, so we don't see the asset loading on the second page
        // if we happened to be swapped to the second page in the main menu
        vgascroll(0);
        vgawait();

        initialize_palette(g_scheme_idx);

        music.reset(&MUSICDEF[hw_cfg.sound_track]);

        // fresh slate of game state
        Field field;
        field.set_level(hw_cfg.starting_level);

        static ScoreDisplay displays[] = {
            { &game_layout.score_1, &SCORE_SCORE, &field.score_display, 0xFFFFFFFF, 1, },
            { &game_layout.score_2, &SCORE_BLOCKS, &field.blocks, 0xFFFFFFFF, 1, },
            { &game_layout.score_3, &SCORE_HIGHSCORE, &save.score, 0xFFFFFFFF, 2, },
            { 0, 0, 0, 0, 0, },
        };

        tileset = 0;

        // load score digits, tiles and bomb animation into second screen page
        prepare_number();
        prepare_cells();
        prepare_graphics();

        // Palette dumped for debugging (visible with vgalinedouble(0) in DOS, via VISIBLE_HEIGHT in SDL)
        // Leave a safe area below the first screenful, so shaking the screen doesn't make this visible
        int safe_y = PORT_HEIGHT+shaking[0]+1;
        for (int co=0; co<256; ++co) {
            putpixel(co, safe_y+1, co);
            putpixel(co, safe_y+2, co);
            putpixel(co, safe_y+3, co);
            putpixel(co, safe_y+4, co);
        }
        putpixel(0, safe_y, 11);
        putpixel(255, safe_y, 11);
        putpixel(0, safe_y+5, 11);
        putpixel(255, safe_y+5, 11);

        time_ms_t last = time_now_ms();

        Timestep ts_time(10UL);
        Timestep ts_palette(50UL);
        Timestep ts_shaking(10UL);

        bool user_quit = false;

        while (!field.quit && !user_quit) {
            time_ms_t now = time_now_ms();
            time_ms_t add = now - last;
            last = now;

            ts_time.update(add);
            ts_palette.update(add);
            ts_shaking.update(add);

            while (ts_time.step()) field.step_time();

            while (ts_palette.step()) palrotate++;

            while (ts_shaking.step()) {
                if (field.shaking_index != -1) {
                    vgascroll(shaking[field.shaking_index++]);
                    if (field.shaking_index == sizeof(shaking) / sizeof(shaking)[0]) {
                        field.shaking_index = -1;
                    }
                }
            }

            PortEvent e;
            while (event_poll(&e)) {
                if (e.type == EVENT_QUIT) {
                    user_quit = true;
                    break;
                }

                if (e.type == EVENT_KEYUP) {
                    if (is_down(e.key)) {
                        field.end_fast_fall();
                    } else if (is_left(e.key) || is_right(e.key)) {
                        field.end_side_movement();
                    }
                }

                if (e.type == EVENT_KEYDOWN) {
                    if (is_left(e.key)) {
                        field.begin_side_movement(-1);
                    } else if (is_right(e.key)) {
                        field.begin_side_movement(+1);
                    } else if (is_down(e.key)) {
                        field.begin_fast_fall();
                    } else if (is_up(e.key) || e.key == KeyCode_n) {
                        field.rotate_block();
                    } else if (is_back(e.key)) {
                        // make sure we reset the shaking scroll
                        vgascroll(0);
                        if (ask_yes_no(QUESTION_END_SESSION)) {
                            user_quit = true;
                        }
                        redraw = true;
                        last = time_now_ms();
                    } else if (is_enter(e.key)) {
                        field.force_fall = true;
                    }
                }
                break;
            }

            if (field.quit || user_quit) {
                break;
            }

            vgawait();

            // redraw was forced or background pattern changed -- redraw
            if (last_diamondmode != diamondmode || redraw) {
                redraw_bg(displays);
                field.redraw_all();
                last_diamondmode = diamondmode;
                redraw = false;
            }

            // render updated scores
            ScoreDisplay *dpy = displays;
            while (dpy->label) {
                if (dpy->last_value != *(dpy->value)) {
                    dpy->last_value = *(dpy->value);
                    diamond(dpy->rect->x, dpy->rect->y+dpy->rect->h-10-ScoreDisplay_inset, dpy->rect->w, 10, diamondmode, 1);
                    render_number(dpy->rect->x+dpy->rect->w-6-ScoreDisplay_inset, dpy->rect->y+dpy->rect->h-10-ScoreDisplay_inset, dpy->last_value, dpy->color);
                }

                ++dpy;
            }

            // render the current game state (intelligent update, only what changed)
            vgapalette(palrotate);
            if (field.cells(render_cell)) {
                // if needed, re-render the slider (vertical line)
                render_vline(game_layout.play_field.x+game_layout.block.w*field.slider_x,
                        game_layout.play_field.y, game_layout.block.h*field.height, 64);
            }
        }

        // new highscore?
        if (field.score > save.score) {
            save.set(field.score);
            save_file(SAVE_FILENAME, &save, sizeof(save));
        }

        if (!user_quit) {
            // force to black and white color scheme, looks like "frozen" at the end of the game
            g_scheme_idx = -1;
            change_colors();
            vgapalette(0);
            // reset scroll
            vgascroll(0);

            render_rect((PORT_WIDTH-8*mystrlen(QUIT_MSG)) / 2 - 11, (PORT_HEIGHT - 8) / 2 - 11, 22+8*mystrlen(QUIT_MSG), 22+8, 1);
            render_rect((PORT_WIDTH-8*mystrlen(QUIT_MSG)) / 2 - 10, (PORT_HEIGHT - 8) / 2 - 10, 20+8*mystrlen(QUIT_MSG), 20+8, 0);
            render_font((PORT_WIDTH-8*mystrlen(QUIT_MSG)) / 2 -1, (PORT_HEIGHT - 8) / 2, QUIT_MSG, 11);
            render_font((PORT_WIDTH-8*mystrlen(QUIT_MSG)) / 2, (PORT_HEIGHT - 8) / 2, QUIT_MSG, 9);

            // This wait-for-one-second-and-eat-all-events is here so that
            // the user actually sees the quit message instead of accidentally
            // pressing a key while still playing the game
            busy_wait_ms(1000);
            PortEvent e;
            // Read out all events
            while (event_poll(&e)) {}

            while (event_wait(&e)) {
                if (e.type == EVENT_KEYDOWN || e.type == EVENT_QUIT) {
                    break;
                }
            }
            vgaclear();

            // back to the default color scheme
            g_scheme_idx = 2;
            initialize_palette(g_scheme_idx);
        }
    }

    loonies_cleanup();
    return 0;
}
