/*
    Loonies 8192 r50 for DOS and SDL platforms
    Copyright (C) 2018-2020 Thomas Perl <m@thp.io>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
*/


#ifndef LOONPORT_H
#define LOONPORT_H

/**
 * Core types, enum definitions and function declarations.
 *
 * loonies_main() is implemented by loonies.cpp (game core)
 * all other functions are implemented by platform/<dos|sdl>/loonport.cpp
 *
 * the function naming isn't consistent, as parts of these
 * functions have been adapted from external library code
 **/

typedef unsigned char uint8_t;
typedef unsigned short uint16_t;
typedef unsigned long time_ms_t;

typedef unsigned long score_t;
// For visual (digit rendering) reasons
#define MAX_SCORE 999999UL
// poor man's "encryption"
#define SCORE_XOR_VALUE 0xBEAABCDBUL

#if defined(__BORLANDC__)
typedef int bool;
#define false 0
#define true 1
#else
#include <stdbool.h>
#endif

#ifdef __cplusplus
extern "C" {
#endif


// Configuration settings

enum Language {
    LANGUAGE_ENGLISH = 0,
    LANGUAGE_GERMAN,
    LANGUAGE_MAX,
};

enum AudioDevice {
    AUDIO_DEV_NONE = 0,
    AUDIO_DEV_PC_SPEAKER,
    AUDIO_DEV_OPL2,
    AUDIO_DEV_MPU401,
    AUDIO_DEV_CDROM,
    AUDIO_DEV_MAX,
};

enum Soundtrack {
    SOUND_TRACK_DESCENT = 0,
    SOUND_TRACK_CLICKEDY,
    SOUND_TRACK_MAX,
};

enum OplSetup {
    AUDIO_OPL_ADLIB_388 = 0,
    AUDIO_OPL_OPL2LPT_LPT1,
    AUDIO_OPL_OPL2LPT_LPT2,
    AUDIO_OPL_MAX,
};

enum MpuSetup {
    AUDIO_MPU_330 = 0,
    AUDIO_MPU_SBMIDI_220,
    AUDIO_MPU_300,
    AUDIO_MPU_310,
    AUDIO_MPU_320,
    AUDIO_MPU_SBMIDI_240,
    AUDIO_MPU_MAX,
};

struct HardwareConfig {
    unsigned char language; // enum Language
    unsigned char audio_dev; // enum AudioDevice
    unsigned char opl_setup; // enum OplSetup
    unsigned char mpu_setup; // enum MpuSetup
    unsigned char sound_track; // enum Soundtrack
    unsigned char starting_level; // 0-10, see gamelogic.h
};


// Event handling

enum PortEventType {
    EVENT_QUIT = 0,
    EVENT_KEYDOWN,
    EVENT_KEYUP,
};

enum KeyCode {
    KeyCode_Left = 0x4B,
    KeyCode_Right = 0x4D,
    KeyCode_Down = 0x50,
    KeyCode_Up = 0x48,
    KeyCode_Enter = 0x1C,
    KeyCode_Escape = 0x01,
    KeyCode_Space = 0x39,
    KeyCode_q = 0x10,
    KeyCode_h = 0x23,
    KeyCode_j = 0x24,
    KeyCode_k = 0x25,
    KeyCode_l = 0x26,
    KeyCode_y = 0x15,
    KeyCode_n = 0x31,
    KeyCode_p = 0x19,
};

struct PortEvent {
    unsigned char type;
    unsigned char key;
};

bool event_poll(struct PortEvent *event);
bool event_wait(struct PortEvent *event);


// Graphics

enum VGAMode {
    VGAMode_Text = 0,
    VGAMode_320x200,
    VGAMode_320x240,
};

void vgamode(enum VGAMode mode);
void putpixel(unsigned short x, unsigned short y, uint8_t color);
unsigned char getpixel(unsigned short x, unsigned short y);
void render_vline(int x, int y, int h, uint8_t color);
void render_rect(int x, int y, int w, int h, uint8_t color);
void vgascroll(unsigned int y);
void vgawait();
void vgaclearbytes(unsigned long bytes);
void vgaclearbytes_y(unsigned short y, long bytes);
void vgablit(unsigned short sx, unsigned short sy,
        unsigned short dx, unsigned short dy,
        unsigned short w, unsigned short h,
        unsigned char orvalue);
void vgablitmemline(int x, int y, const unsigned char *data, int w);
void set_ega_color(uint8_t color, uint8_t r, uint8_t g, uint8_t b);


// Adlib I/O

void WriteFM(unsigned short port, uint8_t reg, uint8_t value);
unsigned short get_lpt_port_base(unsigned char index);


// MIDI I/O

void midi_out(unsigned short port, unsigned char data);
unsigned char midi_in(unsigned short port);


// File I/O

bool save_file(const char *filename, void *data, unsigned short length);
bool load_file(const char *filename, void *data, unsigned short length);


// Timer handling

typedef void (*timer_callback_t)();

void timer_init();
void set_timer_callback(timer_callback_t callback);
time_ms_t time_now_ms();
void timer_quit();


// CD Audio handling

bool cd_available();
void cd_init();
void cd_play_track(int track);
void cd_process();
void cd_stop();


// PC Speaker interfacing

void pcspk(unsigned char enabled, unsigned short frequency);


// Main function

int loonies_main();

// Layout and size
#if defined(__BORLANDC__)
#include "game_layout_dos.h"
#else
#include "game_layout_sdl.h"
#endif

#ifdef __cplusplus
};
#endif

#endif
