/*
    Loonies 8192 r50 for DOS and SDL platforms
    Copyright (C) 2018-2020 Thomas Perl <m@thp.io>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
*/


#include "loonport.h"

#include <dos.h>
#include <io.h>
#include <mem.h>
#include <fcntl.h>

typedef unsigned char far *vga_ptr_t;
vga_ptr_t VGA = (vga_ptr_t)0xA0000000L;

extern "C" {
// some things needed to be set for Borland C when not linking against stdlib
unsigned _stklen = 0x1000;
unsigned _heaplen = 0x0000;
unsigned _nfile = 2;
void _exit(int status) {
    asm {
        mov ax, status
        mov ah, 0x4c
        int 0x21
    }
}
void _npure_error_() { _exit(1); }
void exit(int status) { _exit(status); }
} // extern "C"

// we don't have any C++ object heap
void *operator new(unsigned int) { return 0; }
void operator delete(void near *) { }

bool
save_file(const char *filename, void *data, unsigned short length)
{
    bool written = false;
    unsigned short handle;
    asm {
        mov ah, 0x3c
        mov cx, 0x00
        mov dx, filename
        int 0x21
        jc error
        mov handle, ax

        mov ah, 0x40
        mov bx, handle
        mov cx, length
        mov dx, data
        int 0x21
        jc nowrite
        jmp closefile
    }
nowrite:
    written = false;
closefile:
    asm {
        mov ah, 0x3e
        mov bx, handle
        int 0x21
    }
    return written;

error:
    return false;
}

bool
load_file(const char *filename, void *data, unsigned short len)
{
    bool success = true;
    unsigned short handle;
    asm {
        mov ah, 0x3d
        mov al, 0x00 // read only
        mov dx, filename
        int 0x21
        jc error
        mov handle, ax

        mov ah, 0x3f
        mov bx, handle
        mov cx, len
        mov dx, data
        int 0x21
        jc noread
        jmp closefile
    }
noread:
    success = false;
closefile:
    asm {
        mov ah, 0x3e
        mov bx, handle
        int 0x21
    }
    return success;
error:
    return false;
}

void WriteFM(unsigned short port, uint8_t reg, uint8_t value)
{
    int i;

    if (port == 0x388) {
        outp(port, reg);              /* set up the register  */
        for (i = 0; i < 6; i++) inp(port);  /* wait 12 cycles       */
        outp(port+1, value);               /* write out the value  */
        for (i = 0; i < 35; i++) inp(port+1); /* wait 84 cycles       */
    } else {
        // Assume it's an OPL2LPT port thing
        // https://github.com/pdewacht/adlipt/blob/master/opl2test/OPL2.cpp#L127
#define PP_NOT_STROBE   0x1
#define PP_NOT_AUTOFD   0x2
#define PP_INIT         0x4
#define PP_NOT_SELECT   0x8

        static uint8_t b[] = {
            PP_NOT_SELECT | PP_NOT_STROBE | PP_INIT,
            PP_NOT_SELECT | PP_NOT_STROBE,
            PP_NOT_SELECT | PP_NOT_STROBE | PP_INIT,

            PP_NOT_SELECT | PP_INIT,
            PP_NOT_SELECT,
            PP_NOT_SELECT | PP_INIT
        };

        unsigned short ctrl = port + 2;
        outp(port, reg);
        outp(ctrl, b[0]);
        outp(ctrl, b[1]);
        outp(ctrl, b[2]);
        for (i = 0; i < 6; i++) inp(ctrl);

        outp(port, value);
        outp(ctrl, b[3]);
        outp(ctrl, b[4]);
        outp(ctrl, b[5]);
        for (i = 0; i < 35; i++) inp(ctrl);
    }
}

timer_callback_t
g_timer_callback = 0;

void set_timer_callback(timer_callback_t callback)
{
    disable();
    g_timer_callback = callback;
    enable();
}

typedef void interrupt (far *interrupt_proc_t)(...);

static interrupt_proc_t prev_timer_handler;
//void interrupt (far *prev_timer_handler)(...);
static time_ms_t timer_ticks;
static unsigned char timer_divider;

static void interrupt far timer_handler(...)
{
    ++timer_ticks;
    ++timer_divider;

    if (g_timer_callback) {
        g_timer_callback();
    }

    // http://www3.telus.net/alexander_russell/course/chapter_5.htm
    // Default rate at which we want to call this = 18.2 Hz
    // Our ticks: 1000 Hz / 18.2 Hz ~ 55x faster, so every 55th
    // tick, chain to the previous timer handler
    if (timer_divider % 55 == 0) {
        timer_divider = 0;
        prev_timer_handler(); // apparently "not the best way to chain",
                              // but with _chain_intr() it locks up sometime?
    } else {
        outp(0x20, 0x20);
    }
}

interrupt_proc_t mygetvect(unsigned char number)
{
    unsigned short hiword;
    unsigned short loword;
    asm {
        push es
        mov ah, 0x35 // http://www.stanislavs.org/helppc/int_21-35.html
        mov al, number
        int 0x21
        mov loword, bx
        mov hiword, es
        pop es
    }

    unsigned long value = hiword;
    value <<= 16;
    value |= loword;
    return (interrupt_proc_t)(value);
}

void mysetvect(unsigned char number, interrupt_proc_t proc)
{
    unsigned long value = (unsigned long)proc;
    unsigned short hiword = (value >> 16) & 0xFFFF;
    unsigned short loword = value & 0xFFFF;
    asm {
        push ds
        mov ah, 0x25 // http://www.stanislavs.org/helppc/int_21-25.html
        mov al, number
        mov dx, loword
        mov ds, hiword
        int 0x21
        pop ds
    }
}

// https://en.wikibooks.org/wiki/X86_Assembly/Programmable_Interval_Timer
// https://github.com/pdewacht/adlipt/blob/master/opl2test/timer.cpp
void timer_init()
{
    timer_ticks = 0;
    timer_divider = 0;

    // Set the timer from the default 18.2 Hz to ~ 1000 Hz, so we
    // get a timer interrupt every millisecond, and our time_now_ms()
    // is more correct than the default timer (~ 55 ms granularity)
    // 1193182 / 1000 Hz = 1193 (0x04a9)
    unsigned char lowbyte = 0xa9;
    unsigned char highbyte = 0x04;

    prev_timer_handler = mygetvect(0x08);
    mysetvect(0x08, timer_handler);

    disable();
    outp(0x43, 0x36); // also seen 0x34 somewhere?!
    outp(0x40, lowbyte);
    outp(0x40, highbyte);
    enable();
}

void timer_quit()
{
    disable();
    outp(0x43, 0x36); // also seen 0x34 somewhere?!
    outp(0x40, 0);
    outp(0x40, 0);
    enable();

    mysetvect(0x08, prev_timer_handler);
}

time_ms_t time_now_ms()
{
    time_ms_t result;
    disable();
    result = timer_ticks;
    enable();
    return result;
}

void set_ega_color(unsigned char color, unsigned char r, unsigned char g, unsigned char b)
{
       /* to set everything (www3.telus.net/alexandeR_russel/course/chapter_5.htm
       asm {
           xor al, al
           mov dx, 0x3c8
           out dx, al
           mov cx, 768
           mov si, ptr
           mov dx, 0x3c9
           rep outsb
       }
       */
       outportb(0x03c8, color);
       outportb(0x03c9, r);
       outportb(0x03c9, g);
       outportb(0x03c9, b);
}

// brackeen.com/vga/unchain.html
#define SC_INDEX 0x03c4
#define SC_DATA 0x03c5
#define CRTC_INDEX 0x3d4
#define CRTC_DATA 0x3d5
#define GR_INDEX 0x3ce
#define GR_DATA 0x3cf
#define MISC_OUTPUT 0x3c2

#define GR_WRITEMODE 0x05
#define GR_DATAROTATE 0x03
#define GR_ENABLE_SET_RESET 0x01

// http://www.osdever.net/FreeVGA/vga/seqreg.htm
#define SC_MAP_MASK_REGISTER 0x02

#define CRTC_MAX_SCAN_LINE 0x09
#define MEMORY_MODE 0x04
#define UNDERLINE_LOC 0x14
#define MODE_CONTROL 0x17

static inline void
vga_sequencer(unsigned char index, unsigned char data)
{
    outp(SC_INDEX, index);
    outp(SC_DATA, data);
}

static inline void
vga_crtc(unsigned char index, unsigned char data)
{
    outp(CRTC_INDEX, index);
    outp(CRTC_DATA, data);
}

static inline void
vga_graphreg(unsigned char index, unsigned char data)
{
    outp(GR_INDEX, index);
    outp(GR_DATA, data);
}

static inline uint8_t
vga_get_graphreg(unsigned char index)
{
    // http://www.dvara.net/hk/download/gpbb/gpbb31.pdf
    outp(GR_INDEX, index);
    return inp(GR_DATA);
}

#define VGAOFFSET(x, y) ((y << 6) + (y << 4) + (x >> 2))

void putpixel(unsigned short x, unsigned short y, uint8_t color)
{
    unsigned char plane = x&3;

    /* select the map mask register + write 2^plane */
    /* write 2^plane */
    vga_sequencer(SC_MAP_MASK_REGISTER, 1 << plane);

    VGA[VGAOFFSET(x, y)] = color;
}

unsigned char getpixel(unsigned short x, unsigned short y)
{
    unsigned char plane = x&3;

    // read map select register
    vga_graphreg(0x04, plane);

    return VGA[VGAOFFSET(x, y)];
}

void vgablit(unsigned short sx, unsigned short sy, unsigned short dx, unsigned short dy, unsigned short w, unsigned short h, unsigned char orvalue)
{
    if (w == 0 || h == 0) {
        return;
    }

    int leftcolumn = sx % 4;
    int lc0 = leftcolumn;
    if (leftcolumn != 0) {
        // Example, sx == 7, sx % 4 == 3 -> need to render 1-pixel column (rightmost) at stripe with sx=4
        // Another, sx == 6, sx % 4 == 2 -> need to render 2-pixel column
        // Another, sx == 9, sx % 4 == 1 -> need to render 3-pixel column
        leftcolumn = 4 - leftcolumn;
        if (leftcolumn > w) {
            leftcolumn = w;
        }
    }
    w -= leftcolumn;
    int rightcolumn = w % 4;
    w /= 4; // w is now the number of columns to be blitted

    // set write mode zero
    uint8_t reg = vga_get_graphreg(GR_WRITEMODE);
    reg &= ~3;
    vga_graphreg(GR_WRITEMODE, reg);

    // set logical operation to OR
    // http://www.osdever.net/FreeVGA/vga/graphreg.htm#03
    reg = vga_get_graphreg(GR_DATAROTATE);
    reg |= 0x10;
    reg &= 0xF0;
    vga_graphreg(GR_DATAROTATE, reg);

    // needed?
    //vga_graphreg(GR_ENABLE_SET_RESET, 0x00);

    if (leftcolumn) {
        // 1 -> 0b1000 -> 0x08
        // 2 -> 0b1100 -> 0x0c
        // 3 -> 0b1110 -> 0x0e
        // 4 -> 0b1111 -> 0x0f --> NOT POSSIBLE, done via center blit
        // >>> [hex(0x0f & ~((1<<(4-x))-1)) for x in (1, 2, 3, 4)]
        // ['0x8', '0xc', '0xe', '0xf']

        vga_sequencer(SC_MAP_MASK_REGISTER, 0x0F & ~((1<<(4-leftcolumn))-1));
        unsigned short soff = VGAOFFSET(sx - lc0, sy);
        unsigned short doff = VGAOFFSET(dx - lc0, dy);

        asm {
            push es

            // Set extra segment register to VGA
            mov ax, 0xa000
            mov es, ax

            // Load OR value
            mov dl, orvalue

            // initialize inner loop
            mov cx, h
        }
next_scanline_left:
        asm {
            // Read into latches
            mov si, soff
            mov dh, es:[si]

            // Write with ORing
            mov di, doff
            mov es:[di], dl

            // Forward to next scanline
            add soff, 0x50
            add doff, 0x50

            // loop to next scanline
            dec cx
            jnz next_scanline_left

            pop es
        }
    }

    if (rightcolumn) {
        // 1 -> 0b0001 -> 0x01
        // 2 -> 0b0011 -> 0x03
        // 3 -> 0b0111 -> 0x07
        // 4 -> 0b1111 -> 0x0f --> NOT POSSIBLE, done via center blit
        // >>> [hex((1<<x)-1) for x in (1, 2, 3, 4)]
        // ['0x1', '0x3', '0x7', '0xf']

        vga_sequencer(SC_MAP_MASK_REGISTER, (1<<rightcolumn)-1);
        unsigned short soff = VGAOFFSET(sx + 4*w, sy);
        unsigned short doff = VGAOFFSET(dx + 4*w, dy);

        asm {
            push es

            // Set extra segment register to VGA
            mov ax, 0xa000
            mov es, ax

            // Load OR value
            mov dl, orvalue

            // initialize inner loop
            mov cx, h
        }
next_scanline_right:
        asm {
            // Read into latches
            mov si, soff
            mov dh, es:[si]

            // Write with ORing
            mov di, doff
            mov es:[di], dl

            // Forward to next scanline
            add soff, 0x50
            add doff, 0x50

            // loop to next scanline
            dec cx
            jnz next_scanline_right

            pop es
        }
    }

    if (w > 0) {
        // set write mask to all planes
        vga_sequencer(SC_MAP_MASK_REGISTER, 0x0F);

        unsigned short soff = VGAOFFSET(sx + leftcolumn, sy);
        unsigned short doff = VGAOFFSET(dx + leftcolumn, dy);
        unsigned short scanline_rewind = 0x50*h - 1;

        // center stripes (full 4 px)
        asm {
            push es

            // Set extra segment register to VGA
            mov ax, 0xa000
            mov es, ax

            // Load OR value
            mov dl, orvalue

            // initializer outer loop
            mov bx, w
        }
next_column:
        asm {
            // initialize inner loop
            mov cx, h
        }
next_scanline:
        asm {
            // Read into latches
            mov si, soff
            mov dh, es:[si]

            // Write with ORing
            mov di, doff
            mov es:[di], dl

            // Forward to next scanline
            add soff, 0x50
            add doff, 0x50

            // loop to next scanline
            dec cx
            jnz next_scanline

            // rewind the scanline
            mov ax, scanline_rewind
            sub soff, ax
            sub doff, ax

            // loop to next column
            dec bx
            jnz next_column

            pop es
        }
    }

    // restore normal logical operation
    reg = vga_get_graphreg(GR_DATAROTATE);
    reg &= 0xE0;
    vga_graphreg(GR_DATAROTATE, reg);
}

void render_rect(int x, int y, int w, int h, uint8_t color)
{
    if (w == 0 || h == 0) {
        return;
    }

    int leftcolumn = x % 4;
    int lc0 = leftcolumn;
    if (leftcolumn != 0) {
        leftcolumn = 4 - leftcolumn;
        if (leftcolumn > w) {
            leftcolumn = w;
        }
    }
    w -= leftcolumn;
    int rightcolumn = w % 4;
    w /= 4;

    // set write mode zero
    uint8_t reg = vga_get_graphreg(GR_WRITEMODE);
    reg &= ~3;
    vga_graphreg(GR_WRITEMODE, reg);

    if (leftcolumn) {
        vga_sequencer(SC_MAP_MASK_REGISTER, 0x0F & ~((1<<(4-leftcolumn))-1));
        unsigned short doff = VGAOFFSET(x - lc0, y);

        asm {
            push es

            // Set extra segment register to VGA
            mov ax, 0xa000
            mov es, ax

            // Load color
            mov dl, color

            // initialize inner loop
            mov cx, h
        }
next_scanline_left:
        asm {
            // Write normally
            mov di, doff
            mov es:[di], dl

            // Forward to next scanline
            add doff, 0x50

            // loop to next scanline
            dec cx
            jnz next_scanline_left

            pop es
        }
    }

    if (rightcolumn) {
        vga_sequencer(SC_MAP_MASK_REGISTER, (1<<rightcolumn)-1);
        unsigned short doff = VGAOFFSET(x + leftcolumn + 4*w, y);

        asm {
            push es

            // Set extra segment register to VGA
            mov ax, 0xa000
            mov es, ax

            // Load color value
            mov dl, color

            // initialize inner loop
            mov cx, h
        }
next_scanline_right:
        asm {
            // Write with overwrite
            mov di, doff
            mov es:[di], dl

            // Forward to next scanline
            add doff, 0x50

            // loop to next scanline
            dec cx
            jnz next_scanline_right

            pop es
        }
    }

    if (w > 0) {
        // set write mask to all planes
        vga_sequencer(SC_MAP_MASK_REGISTER, 0x0F);

        unsigned short doff = VGAOFFSET(x + leftcolumn, y);
        unsigned short scanline_rewind = 0x50*h - 1;

        // center stripes (full 4 px)
        asm {
            push es

            // Set extra segment register to VGA
            mov ax, 0xa000
            mov es, ax

            // Load color value
            mov dl, color

            // initializer outer loop
            mov bx, w
        }
next_column:
        asm {
            // initialize inner loop
            mov cx, h
        }
next_scanline:
        asm {
            // Write with ORing
            mov di, doff
            mov es:[di], dl

            // Forward to next scanline
            add doff, 0x50

            // loop to next scanline
            dec cx
            jnz next_scanline

            // rewind the scanline
            mov ax, scanline_rewind
            sub doff, ax

            // loop to next column
            dec bx
            jnz next_column

            pop es
        }
    }
}

void render_vline(int x, int y, int h, uint8_t color)
{
    render_rect(x, y, 1, h, color);
}

void vgablitmemline(int x, int y, const unsigned char *data, int w)
{
    if (x % 4 != 0) {
        int side = 4 - (x % 4);

        for (int i=0; i<side; ++i) {
            putpixel(x+i, y, data[i]);
        }

        x += side;
        data += side;
        w -= side;
    }

    // select map mask register only once
    outp(SC_INDEX, SC_MAP_MASK_REGISTER);
    int offset0 = VGAOFFSET(x, y);
    for (int plane=0; plane<4; ++plane) {
        outp(SC_DATA, 1 << plane);

        vga_ptr_t out = VGA + offset0;
        const unsigned char *input = data + plane;
        for (int i=plane; i<w; i+=4) {
            *out++ = *input;
            input += 4;
        }
    }
}

void vgaclearbytes_y(unsigned short y, long bytes)
{
    // 4 planes at once, 2 bytes per stosw
    bytes /= (4L * 2L);
    unsigned int length = bytes;

    unsigned short off = y * 80;

    // stackoverflow.com/questions/41318704
    // TODO: plane select?
    vga_sequencer(SC_MAP_MASK_REGISTER, 0x0F);
    asm {
        cld
        push es
        mov ax, 0xa000
        mov bx, off
        shr bx, 4
        add ax, bx
        mov es, ax
        mov ax, 0x00
        xor di, di
        mov cx, length
        rep stosw
        pop es
    }
}

void vgaclearbytes(unsigned long bytes)
{
    vgaclearbytes_y(0, bytes);
}

void vgalinedouble(unsigned char count)
{
    // http://www.dvara.net/hk/download/gpbb/gpbb31.pdf
    outp(CRTC_INDEX, CRTC_MAX_SCAN_LINE);
    unsigned char old = inp(CRTC_DATA);
    old &= 0xe0; // set maximum scan line = 0
    old |= count;
    // the low 5 bits can be set to a value, in which
    // case the lines will be doubled (e.g. 1 = 2x doubling)
    outp(CRTC_DATA, old);
}

void vgamode(enum VGAMode mode)
{
    uint8_t bios_mode = (mode == VGAMode_Text) ? 0x03 : 0x13;

    if (bios_mode == 0x13) {
        // clear all 256 KB of vga memory (FIXME: is it 256000 or 256*1024?)
        vgaclearbytes(4L*320L*200L);
    }

    asm {
        mov ah, 0
        mov al, bios_mode
        int 0x10
    }

    if (mode != VGAMode_Text) {
        // turn off chain-4 mode
        vga_sequencer(MEMORY_MODE, 0x06);

        // set mask to all planes
        vga_sequencer(SC_MAP_MASK_REGISTER, 0x0F);

        if (mode == VGAMode_320x240) {
            // http://www.osdever.net/FreeVGA/vga/extreg.htm
            // http://afturgurluk.org/documents/Info/Demo/GraphPro/gp-jul91.asc
            // select 28 MHz dot clock & 60 Hz scanning rate
            vga_sequencer(0x00, 0x01); // sync reset while switching clocks
            outp(MISC_OUTPUT, 0xe7); // select 28 MHz / 60 Hz
            vga_sequencer(0x00, 0x03); // undo reset (restart sequencer)

            outp(CRTC_INDEX, 0x11); // protect bit
            unsigned char value = inp(CRTC_DATA); // ger current vsync end register setting
            value &= 0x7f; // set CRTC Registers Protect Enable to zero
            outp(CRTC_DATA, value); // remove write protect on the vretrace end register

            // http://www.osdever.net/FreeVGA/vga/crtcreg.htm
            vga_crtc(0x06, 0x0d); // vertical total
            vga_crtc(0x07, 0x3e); // overflow (bit 8 of vertical counts)
            // cell height (2 to double-scan) (done later in vgalinedouble)
            vga_crtc(0x10, 0xea); // vsync start
            vga_crtc(0x11, 0xac); // vsync end and protect cr0-cr7
            vga_crtc(0x12, 0xdf); // vertical displayed
            vga_crtc(0x15, 0xe7); // vblank start
            vga_crtc(0x16, 0x06); // vblank end
        }

        vga_crtc(UNDERLINE_LOC, 0x00); // turn off long mode
        vga_crtc(MODE_CONTROL, 0xe3); // turn on byte mode

        // clear all 256 KB of vga memory (FIXME: is it 256000 or 256*1024?)
        vgaclearbytes(4L*320L*200L);

        // If you set this to vgalinedouble(0), then the VGA will display the
        // first two pages (squished), which is useful for debugging purposes
        vgalinedouble(1);
    }
}

/* CRT controller registers */
#define HIGH_ADDRESS 0x0C
#define LOW_ADDRESS  0x0D

unsigned long visible_page = 0;

void vgascroll(unsigned int y)
{
    visible_page = y * 80;
}

void vgawait()
{
    // http://www.brackeen.com/vga/unchain.html#6
    unsigned short high_addr = HIGH_ADDRESS | (visible_page & 0xff00);
    unsigned short low_addr = LOW_ADDRESS | ((visible_page << 8) & 0xff00);

    outpw(CRTC_INDEX, high_addr);
    outpw(CRTC_INDEX, low_addr);

    // github.com/fbergr/dos-mode-13h/blob/master/main.asm
    asm {
        mov dx, 0x3da
    }
wait_vr_start:
    asm {
        in al, dx
        test al, 8
        jnz wait_vr_start
    }

wait_vr_finish:
    asm {
        in al, dx
        test al, 8
        jz wait_vr_finish
    }
}

void midi_out(unsigned short port, unsigned char data)
{
    asm {
        mov dx, port
        mov al, data
        out dx, al
    }
}

unsigned char midi_in(unsigned short port)
{
    unsigned char result;
    asm {
        mov dx, port
        in al, dx
        mov result, al
    }
    return result;
}

unsigned short get_lpt_port_base(unsigned char index)
{
    // https://wiki.osdev.org/Memory_Map_(x86)#BIOS_Data_Area_.28BDA.29
    unsigned short far *lpt_config = (unsigned short far *)0x00400008;
    return lpt_config[index];
}

static void interrupt (far *oldkb)(...); // BIOS keyboard handler

unsigned char scancode;
unsigned char scancode_queue[256];
unsigned char scan_head;
unsigned char scan_tail;
unsigned char keystates[256];

// http://www3.telus.net/alexander_russell/course/chapter_4.htm
// https://wiki.osdev.org/%228042%22_PS/2_Controller#PS.2F2_Controller_IO_Ports
void interrupt
get_scan(...)
{
   asm {
      in al, 0x60 // read scancode
      mov scancode, al
   }

   scancode_queue[scan_tail++] = scancode;
   if (scan_tail == scan_head) {
       // drop events
       ++scan_head;
   }

   asm {
      mov al, 0x20 // reset PIC
      out 0x20, al
   }
}

void init_keyboard()
{
    unsigned char far *bios_key_state = (unsigned char far *)MK_FP(0x40, 0x17);
    // save old BIOS keyboard handler
    oldkb = getvect(9);

    for (int i=0; i<256; ++i) {
        keystates[i] = 0;
    }

    //*bios_key_state &= (~(32 | 64)); // toggle off caps lock and num lock
    //oldkb(); // call bios handler to change keyboard lights

    scan_head = 0;
    scan_tail = 0;
    scancode = 0;

    // install our own handler
    setvect(9, get_scan);
}

void deinit_keyboard()
{
    setvect(9, oldkb);
}

bool event_wait(PortEvent *event)
{
    while (!event_poll(event));
    return true;
}

bool event_poll(PortEvent *event)
{
    unsigned char code;

    disable();
    if (scan_head == scan_tail) {
        enable();
    } else {
        code = scancode_queue[scan_head++];
        enable();

        if (code & 0x80) {
            code &= 0x7F;
            if (keystates[code]) {
                event->type = EVENT_KEYUP;
                event->key = code;
                keystates[code] = 0;
                return true;
            }
        } else {
            if (!keystates[code]) {
                event->type = EVENT_KEYDOWN;
                event->key = code;
                keystates[code] = 1;
                return true;
            }
        }
        event->key = code;
    }

    return false;
}

/*
Based on:
CDROM AUDIO PLAYER By Barry Egerter
Written July 18, 1994; Using Borland C++ 3.1
Code and program: FREEWARE - alter and use at will.
Internet Email:      barry.egerter@softnet.com
*/

#define CDROM 0x21
#define CLOSE_TRAY 5
#define DATA_TRACK 64
#define LOCK 1
#define UNLOCK 0
#define BUSY  512
#define TRACK_MASK 208

typedef struct cdrom_data_t {
  unsigned short drives;
  unsigned char  first_drive;
  unsigned short current_track;
  unsigned long  track_position;
  unsigned char  track_type;
  unsigned char  low_audio;
  unsigned char  high_audio;
  unsigned char  disk_length_min;
  unsigned char  disk_length_sec;
  unsigned char  disk_length_frames;
  unsigned long         endofdisk;
  unsigned char  upc[7];
  unsigned char  diskid[6];
  unsigned long  status;
  unsigned short error;      /* See description below */
} cdrom_data_t;

cdrom_data_t cdrom_data;

void cd_get_audio_info (void);
void cd_set_track (short tracknum);
void cd_seek (unsigned long location);
void cd_play_audio (unsigned long begin, unsigned long end);
void cd_stop_audio (void);
void cd_cmd (unsigned char mode);
short cd_done_play (void);
void cd_lock (unsigned char doormode);

void device_request (void *block)
{
    union REGS inregs, outregs;
    struct SREGS sregs;

    inregs.x.ax = 0x1510;
    inregs.x.cx = cdrom_data.first_drive;
    inregs.x.bx = FP_OFF (block);
    sregs.es = FP_SEG (block);
    int86x (0x2f, &inregs, &outregs, &sregs);
}


void red_book (unsigned long value, unsigned char *min, unsigned char *sec, unsigned char *frame)
{
  *frame = value & 0x000000ff;
  *sec = (value & 0x0000ff00) >> 8;
  *min = (value & 0x00ff0000) >> 16;
}


unsigned long hsg (unsigned long value)
{
  unsigned char min, sec, frame;

  red_book (value, &min, &sec, &frame);
  value = (unsigned long)min * 4500;
  value += (short)sec * 75;
  value += frame - 150;
  return value;
}


void cd_get_audio_info (void)
{
  struct {
    unsigned char length;
    unsigned char subunit;
    unsigned char comcode;
    unsigned short status;
    char ununsed[8];
    unsigned char media;
    long address;
    short bytes;
    short sector;
    long volid;
  } ioctli;

  struct {
    unsigned char mode;
    unsigned char lowest;
    unsigned char highest;
    unsigned long address;
  } track_data;

  ioctli.length = sizeof (ioctli);
  ioctli.subunit = 0;
  ioctli.comcode = 3;
  ioctli.media = 0;
  ioctli.sector = 0;
  ioctli.volid = 0;
  ioctli.address = (long)&track_data;
  ioctli.bytes = sizeof (track_data);
  track_data.mode = 0x0a;
  device_request (&ioctli);
  memcpy (&cdrom_data.diskid, &track_data.lowest, 6);
  cdrom_data.low_audio = track_data.lowest;
  cdrom_data.high_audio = track_data.highest;
  red_book (track_data.address, &cdrom_data.disk_length_min, &cdrom_data.disk_length_sec, &cdrom_data.disk_length_frames);
  cdrom_data.endofdisk = hsg (track_data.address);
  cdrom_data.error = ioctli.status;
}


void cd_set_track (short tracknum)
{
  struct {
    unsigned char length;
    unsigned char subunit;
    unsigned char comcode;
    unsigned short status;
    char ununsed[8];
    unsigned char media;
    unsigned long address;
    unsigned short bytes;
    unsigned short sector;
    unsigned long  volid;
  } tray_request;

  struct {
    unsigned char mode;
    unsigned char track;
    unsigned long address;
    unsigned char control;
  } track_data;

  tray_request.length = sizeof (tray_request);
  tray_request.subunit = 0;
  tray_request.comcode = 3;
  tray_request.media = 0;
  tray_request.media = tray_request.sector = tray_request.volid = 0;
  tray_request.address = (unsigned long)&track_data;
  tray_request.bytes = 7;
  track_data.mode = 0x0b;
  track_data.track = tracknum;
  device_request (&tray_request);
  cdrom_data.error = tray_request.status;
  cdrom_data.track_position = hsg (track_data.address);
  cdrom_data.current_track = tracknum;
  cdrom_data.track_type = track_data.control & TRACK_MASK;
}


void cd_seek (unsigned long location)
{
  unsigned char min, sec, frame;
  struct {
    unsigned char length;
    unsigned char subunit;
    unsigned char comcode;
    unsigned short status;
    char ununsed[8];
    unsigned char addressmode;
    unsigned long transfer;
    unsigned short sectors;
    unsigned long seekpos;
  } play_request;

  play_request.length = sizeof (play_request);
  play_request.subunit = 0;
  play_request.comcode = 131;
  play_request.addressmode = 0;
  play_request.transfer = 0;
  play_request.sectors = 0;
  play_request.seekpos = location;
  device_request (&play_request);
  cdrom_data.error = play_request.status;
}


void cd_play_audio (unsigned long begin, unsigned long end)
{
  unsigned long leng;

  struct {
    unsigned char length;
    unsigned char subunit;
    unsigned char comcode;
    unsigned short status;
    char ununsed[8];
    unsigned char addressmode;
    unsigned long start;
    unsigned long playlength;
  } play_request;

  play_request.length = sizeof (play_request);
  play_request.subunit = 0;
  play_request.comcode = 132;
  play_request.addressmode = 0;
  play_request.start = begin;
  play_request.playlength = end-begin;
  device_request (&play_request);
  cdrom_data.error = play_request.status;
}


void cd_stop_audio (void)
{
  struct {
    unsigned char length;
    unsigned char subunit;
    unsigned char comcode;
    unsigned short status;
    char ununsed[8];
  } stop_request;

  stop_request.length = sizeof (stop_request);
  stop_request.subunit = 0;
  stop_request.comcode = 133;
  device_request (&stop_request);
  cdrom_data.error = stop_request.status;
}


void cd_cmd (unsigned char mode)
{
  struct {
    unsigned char length;
    unsigned char subunit;
    unsigned char comcode;
    unsigned short status;
    char ununsed[8];
    unsigned char media;
    unsigned long address;
    unsigned short bytes;
    unsigned char unused[4];
  } tray_request;

  unsigned char cd_mode;

  cd_mode = mode;
  tray_request.length = sizeof (tray_request);
  tray_request.subunit = 0;
  tray_request.comcode = 12;
  tray_request.media = 0;
  tray_request.address = (unsigned long)&cd_mode;
  tray_request.bytes = 1;
  device_request (&tray_request);
  cdrom_data.error = tray_request.status;
}

short cd_done_play (void)
{
  cd_cmd (CLOSE_TRAY);
  return ((cdrom_data.error & BUSY) == 0);
}

void cd_lock (unsigned char doormode)
{
  struct {
    unsigned char length;
    unsigned char subunit;
    unsigned char comcode;
    unsigned short status;
    char ununsed[8];
    unsigned char media;
    unsigned long address;
    unsigned short bytes;
    unsigned char unused[4];
  } tray_request;

  struct {
    unsigned char mode;
    unsigned char media;
  } cd_data;

  tray_request.length = sizeof (tray_request);
  tray_request.subunit = 0;
  tray_request.comcode = 12;
  tray_request.media = 0;
  tray_request.address = (unsigned long)&cd_data;
  tray_request.bytes = 2;
  cd_data.mode = 1;
  cd_data.media = doormode;
  device_request (&tray_request);
  cdrom_data.error = tray_request.status;
}

static int
g_cd_audio_track = 0;

static int
g_cd_audio_have_cd = 0;

bool cd_available()
{
    union REGS inregs, outregs;

    inregs.h.ah = 0x15;
    inregs.h.al = 0x00;
    inregs.x.bx = 0;
    int86 (0x2f, &inregs, &outregs);
    if (outregs.x.bx == 0)
        return false;
    cdrom_data.drives = outregs.x.bx;
    cdrom_data.first_drive = outregs.x.cx;
    cd_get_audio_info ();
    return true;
}

void cd_init()
{
    short error_ctr = 0;

    do {
        cd_get_audio_info ();
        error_ctr++;
    } while ((cdrom_data.error == 0x8102) && (error_ctr < 500));

    g_cd_audio_have_cd = (cdrom_data.error != 0x8102);
}

void cd_play_track(int track)
{
    if (!g_cd_audio_have_cd) {
        return;
    }

    g_cd_audio_track = track;
    cd_lock(LOCK);

    short temp;
    unsigned long startpos, endpos;

    if ((track < cdrom_data.low_audio) || (track > cdrom_data.high_audio)) {
        return;
    }

    cd_set_track (track);
    if (cdrom_data.track_type == DATA_TRACK) {
        return;
    }

    startpos = cdrom_data.track_position;
    temp = cdrom_data.current_track;
    if ((temp+1) > cdrom_data.high_audio) {
        endpos = cdrom_data.endofdisk;
    } else {
        cd_set_track (temp+1);
        endpos = cdrom_data.track_position;
        cd_set_track (temp);
    }

    cd_seek(startpos);
    cd_play_audio(startpos, endpos);
}

void cd_process()
{
    if (cd_done_play() && g_cd_audio_track) {
        cd_play_track(g_cd_audio_track);
    }
}

void cd_stop()
{
    g_cd_audio_track = 0;
    cd_stop_audio();
    cd_lock(UNLOCK);
}

// http://heim.ifi.uio.no/~inf3150/grupper/1/pcspeaker.html
#define PIT_FREQ 0x1234DD

void pcspk(unsigned char enabled, unsigned short frequency)
{
    if (enabled) {
        outp(0x61, inp(0x61) | 3); // enable speaker
    } else {
        outp(0x61, inp(0x61) & ~3); // disable speaker
    }

    unsigned short counter = (PIT_FREQ / frequency);
    outp(0x42, counter & 0xff); // LSB
    outp(0x42, (unsigned char)(counter >> 8)); // MSB
}

void
main()
{
    init_keyboard();
    int res = loonies_main();
    deinit_keyboard();
    exit(res);
}
